#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
distances and norms for the legacy data module
"""
debug = False 

def Lnorm(weights, p=1, axis=None):
  """calculate L-p norm of weights

Args:
    weights (array(float)): an array of weights
    p (int, default=1): the power of the p-norm, where ``p in [0,inf]``
    axis (int, default=None): axis used to take the norm along

Returns:
    a float distance norm for the weights
"""
  from numpy import asarray, seterr, inf, abs, max, sum, expand_dims
  weights = asarray(weights, dtype=float)
  if not p:
    w = sum(weights != 0.0, dtype=float, axis=axis) # number of nonzero elements
  elif p == inf:
    w = max(abs(weights), axis=axis)
  elif p == -inf: #XXX: special case, as p in [0,inf]
    w = min(abs(weights), axis=axis)
  else:
    orig = seterr(over='raise', invalid='raise')
    try:
      w = sum(abs(weights**p), axis=axis)**(1./p)
    except FloatingPointError: # use the infinity norm
      w = max(abs(weights), axis=axis)
    seterr(**orig)
  return w if (axis is None or not w.shape) else expand_dims(w, axis=axis)

def absolute_distance(x, xp=None, pair=False, dmin=0):
  """pointwise (or pairwise) absolute distance

``pointwise = |x.T[:,newaxis] - x'.T|  or  pairwise = |x.T - x'.T|.T``

Args:
    x (array): an array of points, ``x``
    xp (array, default=None): a second array of points, ``x'``
    pair (bool, default=False): if True, return the pairwise distances
    dmin (int, default=0): upconvert ``x,x'`` to ``dimension >= dmin``

Returns:
    an array of absolute distances between points

Notes:
    - ``x'==x`` is symmetric with zeros on the diagonal 
    - use ``dmin=2`` for the forced upconversion of 1-D arrays
"""
  from numpy import abs, asarray, newaxis as nwxs, zeros_like
  # cast as arrays of the same dimension
  x = asarray(x)
  xp = x if xp is None else asarray(xp)
  xsize = max(len(x.shape), len(xp.shape), dmin)
  while len(x.shape) < xsize: x = x[nwxs]
  while len(xp.shape) < xsize: xp = xp[nwxs]
  # prep for build manhattan matrix in single operation
  if pair:
    return abs(x.T - xp.T).T
  xsl = (slice(None),)*xsize + (None,)           #NOTE: [:,:,nwxs] for 2-D
  xpsl = (slice(None),)*max(0,xsize-1) + (None,) #NOTE: [:,nwxs] for 2-D
  return abs(x.T[xsl] - xp.T[xpsl])


def lipschitz_metric(L, x, xp=None):
  """sum of lipschitz-weighted distance between points

``d = sum(L[i] * |x[i] - x'[i]|)``

Args:
    L (array): an array of Lipschitz constants, ``L``
    x (array): an array of points, ``x``
    xp (array, default=None): a second array of points, ``x'``

Returns:
    an array of absolute distances between points
"""
  #FIXME: merge with lipschitz cone (distance/contains)
  from numpy import sum, asarray
  d = absolute_distance(x,xp) #XXX: ,dmin=2)
  return sum(L * d.T, axis=-1).T

def _npts(*x):
  """get len(product measure), given lengths of each underlying measure"""
  from numpy import product
  return product(x)

def _get_xy(points):
  """extract the list of positions and the list of values for given points"""
  from numpy import asarray
  if hasattr(points, 'coords'):
    x  = points.coords
    y  = points.values
  elif len(asarray(points).shape) >= 1:
    x = [p.position for p in points] 
    y = [p.value for p in points] 
  else:
    x = points.position
    y = points.value
  return x,y

###########################################################################
# distance metrics
###########################################################################

def chebyshev(x,xp=None, pair=False, dmin=0, axis=None):
  """infinity norm distance between points in euclidean space

``d(inf) =  max(|x[0] - x[0]'|, |x[1] - x[1]'|, ..., |x[n] - x[n]'|)``

Args:
    x (array): an array of points, ``x``
    xp (array, default=None): a second array of points, ``x'``
    pair (bool, default=False): if True, return the pairwise distances
    dmin (int, default=0): upconvert ``x,x'`` to ``dimension >= dmin``
    axis (int, default=None): if not None, reduce across the given axis

Returns:
    an array of absolute distances between points

Notes:
    most common usage has ``pair=False`` and ``axis=0``, or
    pairwise distance with ``pair=True`` and ``axis=1``
"""
  d = absolute_distance(x,xp,pair=pair,dmin=dmin)
  return d.max(axis=axis).astype(float)


def hamming(x,xp=None, pair=False, dmin=0, axis=None):
  """zero 'norm' distance between points in euclidean space

``d(0) =  sum(x[0] != x[0]', x[1] != x[1]', ..., x[n] != x[n]')``

Args:
    x (array): an array of points, ``x``
    xp (array, default=None): a second array of points, ``x'``
    pair (bool, default=False): if True, return the pairwise distances
    dmin (int, default=0): upconvert ``x,x'`` to ``dimension >= dmin``
    axis (int, default=None): if not None, reduce across the given axis

Returns:
    an array of absolute distances between points

Notes:
    most common usage has ``pair=False`` and ``axis=0``, or
    pairwise distance with ``pair=True`` and ``axis=1``
"""
  d = absolute_distance(x,xp,pair=pair,dmin=dmin)
  return d.astype(bool).sum(axis=axis).astype(float)


def minkowski(x,xp=None, pair=False, dmin=0, p=3, axis=None):
  """p-norm distance between points in euclidean space

``d(p) = sum(|x[0] - x[0]'|^p, |x[1] - x[1]'|^p, ..., |x[n] - x[n]'|^p)^(1/p)``

Args:
    x (array): an array of points, ``x``
    xp (array, default=None): a second array of points, ``x'``
    pair (bool, default=False): if True, return the pairwise distances
    dmin (int, default=0): upconvert ``x,x'`` to ``dimension >= dmin``
    p (int, default=3): value of p for the p-norm
    axis (int, default=None): if not None, reduce across the given axis

Returns:
    an array of absolute distances between points

Notes:
    most common usage has ``pair=False`` and ``axis=0``, or
    pairwise distance with ``pair=True`` and ``axis=1``
"""
  from numpy import seterr, inf
  if p == inf: return chebyshev(x,xp,pair=pair,dmin=dmin,axis=axis)
  d = absolute_distance(x,xp,pair=pair,dmin=dmin)
  orig = seterr(over='raise', invalid='raise')
  try:
      d = (d**p).sum(axis=axis)**(1./p)
  except FloatingPointError: # use the infinity norm
      d = d.max(axis=axis).astype(float)
  seterr(**orig)
  return d


def euclidean(x,xp=None, pair=False, dmin=0, axis=None):
  """L-2 norm distance between points in euclidean space

``d(2) = sqrt(sum(|x[0] - x[0]'|^2, |x[1] - x[1]'|^2, ..., |x[n] - x[n]'|^2))``

Args:
    x (array): an array of points, ``x``
    xp (array, default=None): a second array of points, ``x'``
    pair (bool, default=False): if True, return the pairwise distances
    dmin (int, default=0): upconvert ``x,x'`` to ``dimension >= dmin``
    axis (int, default=None): if not None, reduce across the given axis

Returns:
    an array of absolute distances between points

Notes:
    most common usage has ``pair=False`` and ``axis=0``, or
    pairwise distance with ``pair=True`` and ``axis=1``
"""
  return minkowski(x,xp,pair=pair,dmin=dmin,p=2,axis=axis)


def manhattan(x,xp=None, pair=False, dmin=0, axis=None):
  """L-1 norm distance between points in euclidean space

``d(1) = sum(|x[0] - x[0]'|, |x[1] - x[1]'|, ..., |x[n] - x[n]'|)``

Args:
    x (array): an array of points, ``x``
    xp (array, default=None): a second array of points, ``x'``
    pair (bool, default=False): if True, return the pairwise distances
    dmin (int, default=0): upconvert ``x,x'`` to ``dimension >= dmin``
    axis (int, default=None): if not None, reduce across the given axis

Returns:
    an array of absolute distances between points

Notes:
    most common usage has ``pair=False`` and ``axis=0``, or
    pairwise distance with ``pair=True`` and ``axis=1``
"""
  return minkowski(x,xp,pair=pair,dmin=dmin,p=1,axis=axis)

###########################################################################

def is_feasible(distance, cutoff=0.0):
  """determine if the distance exceeds the given cutoff distance

Args:
    distance (array): the measure of feasibility for each point
    cutoff (float, default=0.0): maximum acceptable distance

Returns:
    bool array, with True where the distance is less than cutoff
"""
  from numpy import asarray
  d = infeasibility(distance, cutoff) > 0.0
  if len(asarray(d).shape) == 0:
    return not d
  return ~d


def infeasibility(distance, cutoff=0.0): 
  """amount by which the distance exceeds the given cutoff distance

Args:
    distance (array): the measure of feasibility for each point
    cutoff (float, default=0.0): maximum acceptable distance

Returns:
    an array of distances by which each point is infeasbile
"""
  from numpy import array
  distance = array(distance)
  # zero-out all distances less than tolerated
  if cutoff is not None:
    if len(distance.shape) == 0:
      return 0.0 if distance <= cutoff else distance
    distance[distance <= cutoff] = 0.0
  return distance


def lipschitz_distance(L, points1, points2, **kwds):
  """calculate the lipschitz distance between two sets of datapoints

Args:
    L (list): a list of lipschitz constants
    points1 (mystic.math.legacydata.dataset): a dataset
    points2 (mystic.math.legacydata.dataset): a second dataset
    tol (float, default=0.0): maximum acceptable deviation from shortness
    cutoff (float, default=tol): zero out distances less than cutoff

Returns:
    a list of lipschitz distances

Notes:
    Both *points1* and *points2* can be a ``mystic.math.legacydata.dataset``,
    or a list of ``mystic.math.legacydata.datapoint`` objects, or a list of
    ``lipschitzcone.vertex`` objects (from ``mystic.math.legacydata``).
    *cutoff* takes a float or a boolean, where ``cutoff=True`` will set the
    value of *cutoff* to the default. Typically, the value of *cutoff* is
    *tol*, 0.0, or None.

    Each point x,y can be thought to have an associated double-cone with slope
    equal to the lipschitz constant. Shortness with respect to another point is
    defined by the first point not being inside the cone of the second. We can
    allow for some error in shortness, a short tolerance *tol*, for which the
    point x,y is some acceptable y-distance inside the cone. While very tightly
    related, *cutoff* and *tol* play distinct roles; *tol* is subtracted from
    calculation of the lipschitz_distance, while *cutoff* zeros out the value
    of any element less than the *cutoff*.
"""
  #FIXME: merge with lipschitz cone (distance/contains)
  x,y   = _get_xy(points1)
  xp,yp = _get_xy(points2)

  # get tolerance in y
  tol = kwds.pop('tol', 0.0)
  cutoff = tol  # default is to zero out distances less than tolerance
  if 'cutoff' in kwds: cutoff = kwds.pop('cutoff')
  if cutoff is True: cutoff = tol
  elif cutoff is False: cutoff = None

  # calculate the distance matrix
  md = absolute_distance(y,yp) - max(0.0, tol)
  lm = lipschitz_metric(L,x,xp)
  d = md - lm
  # zero-out all distances less than tolerated 
  return infeasibility(d, cutoff)


def graphical_distance(model, points, **kwds):
  """find the ``radius(x')`` that minimizes the graph between reality (data),
``y = G(x)``, and an approximating function, ``y' = F(x')``.

Args:
    model (func): a model ``y' = F(x')`` that approximates reality ``y = G(x)``
    points (mystic.math.legacydata.dataset): a dataset, defines ``y = G(x)``
    ytol (float, default=0.0): maximum acceptable difference ``|y - F(x')|``.
    xtol (float, default=0.0): maximum acceptable difference ``|x - x'|``.
    cutoff (float, default=ytol): zero out distances less than cutoff.
    hausdorff (bool, default=False): hausdorff ``norm``, where if given,
        then ``ytol = |y - F(x')| + |x - x'|/norm``.

Returns:
    the radius (the minimum distance ``x,G(x)`` to ``x',F(x')`` for each ``x``)

Notes:
    *points* can be a ``mystic.math.legacydata.dataset`` or a list of
    ``mystic.math.legacydata.datapoint`` objects.

    *xtol* defines the n-dimensional base of a pilar of height *ytol*,
    centered at each point. The region inside the pilar defines the space
    where a "valid" model must intersect. If *xtol* is not specified, then
    the base of the pilar will be a dirac at ``x' = x``. This function
    performs an optimization for each ``x`` to find an appropriate ``x'``.

    *ytol* is a single value, while *xtol* is a single value or an iterable.
    *cutoff* takes a float or a boolean, where ``cutoff=True`` will set the
    value of *cutoff* to the default. Typically, the value of *cutoff* is
    *ytol*, 0.0, or None. *hausdorff* can be False (e.g. ``norm = 1.0``),
    True (e.g. ``norm = spread(x)``), or a list of points of ``len(x)``.

    While *cutoff* and *ytol* are very tightly related, they play a distinct
    role; *ytol* is used to set the optimization termination for an acceptable
    ``|y - F(x')|``, while *cutoff* is applied post-optimization.

    If we are using the hausdorff norm, then *ytol* will set the optimization
    termination for an acceptable ``|y - F(x')| + |x - x'|/norm``, where the
    ``x`` values are normalized by ``norm = hausdorff``.
""" #FIXME: update docs to show normalization in y
 #NotImplemented:
 #L = list of lipschitz constants, for use when lipschitz metric is desired
 #constraints = constraints function for finding minimum distance
  from mystic.math.legacydata import dataset
  from numpy import asarray, sum, isfinite, zeros, seterr
  from mystic.solvers import diffev2, fmin_powell
  from mystic.monitors import Monitor, VerboseMonitor

  # ensure target xe and ye is a dataset
  target = dataset()
  target.load(*_get_xy(points))
  nyi = target.npts             # y's are target.values
  nxi = len(target.coords[-1])  # nxi = len(x) / len(y)
  
  # NOTE: the constraints function is a function over a single xe,ye
  #       because each underlying optimization is over a single xe,ye.
  #       thus, we 'pass' on using constraints at this time...
  constraints = None   # default is no constraints
  if 'constraints' in kwds: constraints = kwds.pop('constraints')
  if not constraints:  # if None (default), there are no constraints
    constraints = lambda x: x

  # get tolerance in y and wiggle room in x
  ytol = kwds.pop('ytol', 0.0)
  xtol = kwds.pop('xtol', 0.0) # default is to not allow 'wiggle room' in x 

  cutoff = ytol  # default is to zero out distances less than tolerance
  if 'cutoff' in kwds: cutoff = kwds.pop('cutoff')
  if cutoff is True: cutoff = ytol
  elif cutoff is False: cutoff = None
  ipop = kwds.pop('ipop', min(20, 3*nxi)) #XXX: tune ipop?
  imax = kwds.pop('imax', 1000) #XXX: tune imax?

  # get range for the dataset (normalization for hausdorff distance)
  hausdorff = kwds.pop('hausdorff', False)
  if not hausdorff:  # False, (), None, ...
    ptp = [0.0]*nxi
    yptp = 1.0
  elif hausdorff is True:
    from mystic.math.measures import spread
    ptp = [spread(xi) for xi in zip(*target.coords)]
    yptp = spread(target.values)  #XXX: this can lead to bad bad things...
  else:
    try: #iterables
      if len(hausdorff) < nxi+1:
        hausdorff = list(hausdorff) + [0.0]*(nxi - len(hausdorff)) + [1.0]
      ptp = hausdorff[:-1]  # all the x
      yptp = hausdorff[-1]  # just the y
    except TypeError: #non-iterables
      ptp = [hausdorff]*nxi
      yptp = hausdorff

  #########################################################################
  def radius(model, point, ytol=0.0, xtol=0.0, ipop=None, imax=None):
    """graphical distance between a single point x,y and a model F(x')"""
    # given a single point x,y: find the radius = |y - F(x')| + delta
    # radius is just a minimization over x' of |y - F(x')| + delta
    # where we apply a constraints function (of box constraints) of
    # |x - x'| <= xtol  (for each i in x)
    #
    # if hausdorff = some iterable, delta = |x - x'|/hausdorff
    # if hausdorff = True, delta = |x - x'|/spread(x); using the dataset range
    # if hausdorff = False, delta = 0.0
    #
    # if ipop, then DE else Powell; ytol is used in VTR(ytol)
    # and will terminate when cost <= ytol
    x,y = _get_xy(point)
    y = asarray(y)
    # catch cases where yptp or y will cause issues in normalization
   #if not isfinite(yptp): return 0.0 #FIXME: correct?  shouldn't happen
   #if yptp == 0: from numpy import inf; return inf #FIXME: this is bad

    # build the cost function
    if hausdorff: # distance in all directions
      def cost(rv):
        '''cost = |y - F(x')| + |x - x'| for each x,y (point in dataset)'''
        _y = model(rv)
        if not isfinite(_y): return abs(_y)
        errs = seterr(invalid='ignore', divide='ignore') # turn off warning 
        z = abs((asarray(x) - rv)/ptp)  # normalize by range
        m = abs(y - _y)/yptp            # normalize by range
        seterr(invalid=errs['invalid'], divide=errs['divide']) # turn on warning
        return m + sum(z[isfinite(z)])
    else:  # vertical distance only
      def cost(rv):
        '''cost = |y - F(x')| for each x,y (point in dataset)'''
        return abs(y - model(rv))

    if debug:
      print("rv: %s" % str(x))
      print("cost: %s" % cost(x))

    # if xtol=0, radius is difference in x,y and x,F(x); skip the optimization
    try:
      if not imax or not max(xtol): #iterables
        return cost(x)
    except TypeError:
      if not xtol: #non-iterables
        return cost(x)

    # set the range constraints
    xtol = asarray(xtol)
    bounds = list(zip( x - xtol, x + xtol ))

    if debug:
      print("lower: %s" % str(zip(*bounds)[0]))
      print("upper: %s" % str(zip(*bounds)[1]))

    # optimize where initially x' = x
    stepmon = Monitor()
    if debug: stepmon = VerboseMonitor(1)
    #XXX: edit settings?
    MINMAX = 1 #XXX: confirm MINMAX=1 is minimization
    ftol = ytol
    gtol = None  # use VTRCOG
    if ipop:
      results = diffev2(cost, bounds, ipop, ftol=ftol, gtol=gtol, \
                        itermon = stepmon, maxiter=imax, bounds=bounds, \
                        full_output=1, disp=0, handler=False)
    else:
      results = fmin_powell(cost, x, ftol=ftol, gtol=gtol, \
                            itermon = stepmon, maxiter=imax, bounds=bounds, \
                            full_output=1, disp=0, handler=False)
   #solved = results[0]            # x'
    func_opt = MINMAX * results[1] # cost(x')
    if debug:
      print("solved: %s" % results[0])
      print("cost: %s" % func_opt)

    # get the minimum distance |y - F(x')|
    return func_opt
   #return results[0], func_opt
  #########################################################################

  #XXX: better to do a single optimization rather than for each point ???
  d = [radius(model, point, ytol, xtol, ipop, imax) for point in target]
  return infeasibility(d, cutoff)


#def split_xy(params, npts):
#  """split params_{w,x,y} to params_{wx}, params_{y}
#  npts is [len(measure1),...,len(measureN)]  i.e. pm.pts"""
# #if not isinstance(npts,int):
#  _len = 2 * sum(npts)
#  return params[:_len], params[_len:]  
# #return params[:-npts], params[-npts:]

#def mend_xy(wx_param, y_param):
#  """append params_{y} to params_{wx} yielding params_{w,x,y}"""
#  param = wx_param[:]
#  param.extend(y_param)
#  return param


if __name__ == '__main__':
  ### conversions ###
  print("building a list of params(w,x,Y)...")
  pts = (2,2,2)
  param1 = [.5,.5,1,2, .25,.75,3,4, .125,.875,5,6, -1,-2,-3,-4,-5,-6,-7,-8]
  print("pts: %s" % str(pts))
  print("params: %s" % param1)
 
  print("\nbuilding a scenario from the params...")
  # [store Y as 'values' OR register(F) for Y=F(X) OR points store y as 'val' ?]
  from mystic.math.discrete import scenario
  pm = scenario()
  pm.load(param1, pts)
  print("pm.wts: %s" % str(pm.wts))
  print("pm.pos: %s" % str(pm.pos))
  W = pm.weights
  X = pm.coords
  Y = pm.values
  print("pm.weights: %s" % str(W))
  print("pm.coords: %s" % str(X))
  print("pm.values: %s" % str(Y))

  print("\nbuilding a dataset from the scenario...")
  # build a dataset (using X,Y)
  # [store W as 'weights' ?]
  from mystic.math.legacydata import dataset
  d = dataset()
  d.load(X, Y) 
  print("d.coords: %s" % str(d.coords))
  print("d.values: %s" % str(d.values))

  print("\nedit the dataset...")
  d[0].value = 0
  print("d.values: %s" % str(d.values))
  # DON'T EDIT d[0].position... IT BREAKS PRODUCT MEASURE!!!

  print("\nupdate the scenario from the dataset...")
  # update pm(w,x,Y | W,X) from dataset(X,Y)
  pm.coords, pm.values = d.fetch()
  print("then, build a new list of params from the scenario...")
  # convert pm(w,x,Y | W,X) to params(w,x,Y)
  param1 = pm.flatten(all=True)
  print("params: %s" % str(param1))

  ### lipschitz ###
  param2 = [1.,0.,1,1, .75,.25,3,4, .5,.5,1,2, -8,-2,4,4,-3,-8,-8,0]
  qm = scenario()
  qm.load(param2, pts)
  L = [.25,.5,1.]

  print("")
 #print("param1: %s" % param1)
 #print("param2: %s" % param2)
  print("pm1.coords: %s" % pm.coords)
  print("pm2.coords: %s" % qm.coords)
  print("L = %s" % L)

  print("manhattan distance...")
  print("pm1[0:3],pm2[0:3] =>\n%s\n" % absolute_distance(pm.coords[0:3],qm.coords[0:3]))
  print("pm1[0:1],pm2[0:3] =>\n%s\n" % absolute_distance(pm.coords[0:1],qm.coords[0:3]))
  print("pm1[0:1],pm2[0:1] =>\n%s\n" % absolute_distance(pm.coords[0:1],qm.coords[0:1]))

  print("pm1[0:3],pm2[0:3] values =>\n%s\n" % absolute_distance(pm.values[0:3],qm.values[0:3]))
  print("pm1[0:1],pm2[0:3] values =>\n%s\n" % absolute_distance(pm.values[0:1],qm.values[0:3]))
  print("pm1[0:1],pm2[0:1] values =>\n%s\n" % absolute_distance(pm.values[0:1],qm.values[0:1]))
   
  print("lipschitz metric...")
  print("pm1[0:3],pm2[0:3] =>\n%s\n" % lipschitz_metric(L, pm.coords[0:3], qm.coords[0:3]))
  print("pm1[0:1],pm2[0:3] =>\n%s\n" % lipschitz_metric(L, pm.coords[0:1], qm.coords[0:3]))
  print("pm1[0;1],pm2[0:1] =>\n%s\n" % lipschitz_metric(L, pm.coords[0:1], qm.coords[0:1]))
  
  print("lipschitz distance...")
  print("(don't cutoff):\n%s" % lipschitz_distance(L, pm, qm, cutoff=None))
  print("from measures:\n%s" % lipschitz_distance(L, pm, qm))
  b = dataset()
  id = ['A','B','C','D','E','F','G','H']
  b.load(qm.coords, qm.values, ids=id) 
  print("from datasets:\n%s" % lipschitz_distance(L, d, b))
  print("from list of points:\n%s" % lipschitz_distance(L, d.raw, b.raw))
  print("individual points:\n%s" % lipschitz_distance(L, d.raw[0:1], b.raw[0:1]))

  print("")
  print("is short:\n%s" % is_feasible( lipschitz_distance(L, pm, qm) ))

  ### updates ###
  print("\nupdates to dataset...")
  print("original:\n%s" % b)
  b.update(qm.coords, qm.values) 
  print("points from pm2:\n%s" % b)
  b.update(pm.coords, pm.values) 
  print("points from pm1:\n%s" % b)

  print("\nupdates to product measure...")
  print("orig: %s\n %s\n from %s" % (pm, pm.values, pm.flatten(all=True)))
  par = [.25,.75,3,4, .5,.5,1,2, .125,.875,5,6, 0,-1,-2,-3,-4,-5,-6,-7]
  pm.update(par)
  print("alt: %s\n %s\n from %s" % (pm, pm.values, par))
  par = [.25,.75,3,4, .5,.5,1,2, .125,.875,5,6, -1,-2,-3,-4,-5,-6,-7,-8,-9]
  pm.update(par)
  print("alt2: %s\n %s\n from %s" % (pm, pm.values, par))
  par = [.25,.75,3,4, .5,.5,1,2, .125,.875,5,6, -2,-4,-6,-8]
  pm.update(par)
  print("alt3: %s\n %s\n from %s" % (pm, pm.values, par))
  par = [.5,.5,1,2, .25,.75,3,4]
  pm.update(par)
  print("alt4: %s\n %s\n from %s" % (pm, pm.values, par))

  ##### indexing #####
  b.update(pm.coords, pm.values)
  assert b[0].value == pm.values[0]
  assert b[1].value == pm.values[1]
  assert b[0].position == pm.coords[0] # slow
  assert b[0].position == pm.select(0, reduce=False)[0] # fast
  assert b[1].position == pm.coords[1] # slow
  assert b[1].position == pm.select(1) # fast
  assert pm.select(0,2) == [pm.select(0),pm.select(2)]

  # member calls #
  print("\ntesting mean_value...")
  print("mean_value: %s" % pm.mean_value())
  pm.set_mean_value(5.0)
  print("mean_value: %s" % pm.mean_value())

  print("\ntesting shortness, feasibility, validity...")
  assert pm.short_wrt_data(b) == True
  b.lipschitz = L
  assert pm.short_wrt_self(L) == False
  assert pm.short_wrt_data(b) == False
  pm.set_feasible(b)
  assert pm.short_wrt_self(L) == True
  assert pm.short_wrt_data(b) == True

  Cy = 0.1; Cx = 0.0
  model = lambda x:x[0]
  assert b.valid(model, ytol=Cy, xtol=Cx) == False
  assert pm.valid_wrt_model(model, ytol=Cy, xtol=Cx) == False
  pm.set_valid(model, cutoff=Cy, xtol=Cx)
  assert pm.valid_wrt_model(model, ytol=Cy, xtol=Cx) == True
  print("...done\n")


# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Defaults for mapper and launcher. These should be available as a minimal
(dependency-free) pure-python install from ``pathos``::

    serial_launcher -- syntax for standard python execution
    python_map      -- wrapper around the standard python map
    worker_pool     -- the worker_pool map strategy
"""

import os
_pid = '.' + str(os.getpid()) + '.'
defaults = {
    'nodes' : '1',
    'program' : '',
    'python' : '`which python`' ,
    'progargs' : '',

    'outfile' : 'results%sout' % _pid,
    'errfile' : 'errors%sout' % _pid,
    'jobfile' : 'job%sid' % _pid,

    'scheduler' : '',
    'timelimit' : '00:02',
    'queue' : '',

    'workdir' : '.'
    }

def serial_launcher(kdict={}):
    """
prepare launch for standard execution
syntax:  (python) (program) (progargs)

Notes:
    run non-python shell commands by setting ``python`` to a null string: 
    ``kdict = {'python':'', ...}``
    """
    mydict = defaults.copy()
    mydict.update(kdict)
    str = """ %(python)s %(program)s %(progargs)s""" % mydict
    return str

def python_map(func, *arglist, **kwds):
    """maps function *func* across arguments *arglist*.

Provides the standard python map function, however also accepts *kwds* in
order to conform with the (deprecated) ``pyina.ez_map`` interface.

Notes:
    The following *kwds* used in ``ez_map`` are accepted, but disabled:
        * nodes -- the number of parallel nodes
        * launcher -- the launcher object
        * scheduler -- the scheduler object
        * mapper -- the mapper object
        * timelimit -- string representation of maximum run time (e.g. '00:02')
        * queue -- string name of selected queue (e.g. 'normal')
"""
   #print("ignoring: %s" % kwds)  #XXX: should allow use of **kwds
    result = list(map(func, *arglist)) #     see pathos.pyina.ez_map
    return result

def worker_pool():
    """use the 'worker pool' strategy; hence one job is allocated to each
worker, and the next new work item is provided when a node completes its work
"""
    #from mpi_pool import parallel_map as map
    #return map
    return "mpi_pool"

# backward compatibility
carddealer_mapper = worker_pool

del os


if __name__=='__main__':
    f = lambda x:x**2
    print(python_map(f,range(5),nodes=10))

    import subprocess
    d = {'progargs': """-c "print('hello')" """}
    subprocess.call(serial_launcher(d), shell=True)


# End of file



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Author: Patrick Hung (patrickh @caltech)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Bevington & Robinson's model of dual exponential decay

References:
    1. "Data Reduction and Error Analysis for the Physical Sciences",
       Bevington & Robinson, Second Edition, McGraw-Hill, New York (1992).
"""
from .abstract_model import AbstractModel

from numpy import array, asarray
from numpy import sum as numpysum
from numpy import exp, sqrt
from mystic.forward_model import CostFactory as CF

class BevingtonDecay(AbstractModel):
    """
Computes dual exponential decay [1].
y = a1 + a2 Exp[-t / a4] + a3 Exp[-t/a5]
    """

    def __init__(self,name='decay',metric=lambda x: numpysum(x*x)):
        AbstractModel.__init__(self,name,metric)
        return

    def evaluate(self,coeffs,evalpts):
        """evaluate dual exponential decay with given coeffs over given evalpts
coeffs = (a1,a2,a3,a4,a5)"""
        a1,a2,a3,a4,a5 = coeffs
        t = asarray(evalpts) #XXX: requires a numpy.array
        return a1 + a2*exp(-t/a4) + a3*exp(-t/a5)

    def ForwardFactory(self,coeffs):
        """generates a dual decay model instance from a list of coefficients"""
        a1,a2,a3,a4,a5 = coeffs
        def forward_decay(evalpts):
            """a dual exponential decay over a 1D numpy array
with (a1,a2,a3,a4,a5) = (%s,%s,%s,%s,%s)""" % (a1,a2,a3,a4,a5)
            return self.evaluate((a1,a2,a3,a4,a5),evalpts)
        return forward_decay

    def CostFactory(self,target,pts):
        """generates a cost function instance from list of coefficients & evaluation points"""
        datapts = self.evaluate(target,pts)
        F = CF()
        F.addModel(self.ForwardFactory,len(target),self.__name__)
        self.__cost__ = F.getCostFunction(evalpts=pts,observations=datapts,sigma=sqrt(datapts),metric=self.__metric__)
        return self.__cost__

    def CostFactory2(self,pts,datapts,nparams):
        """generates a cost function instance from datapoints & evaluation points"""
        F = CF()
        F.addModel(self.ForwardFactory,nparams,self.__name__)
        self.__cost__ = F.getCostFunction(evalpts=pts,observations=datapts,sigma=sqrt(datapts),metric=self.__metric__)
        return self.__cost__

    pass
 

# prepared instances
decay = BevingtonDecay() #FIXME: look up the correct name for the model!

# data from Chapter 8 of [1].
data = array([[15, 775], [30, 479], [45, 380], [60, 302],
[75, 185], [90, 157], [105,137], [120, 119], [135, 110],
[150, 89], [165, 74], [180, 61], [195, 66], [210, 68],
[225, 48], [240, 54], [255, 51], [270, 46], [285, 55],
[300, 29], [315, 28], [330, 37], [345, 49], [360, 26],
[375, 35], [390, 29], [405, 31], [420, 24], [435, 25],
[450, 35], [465, 24], [480, 30], [495, 26], [510, 28],
[525, 21], [540, 18], [555, 20], [570, 27], [585, 17],
[600, 17], [615, 14], [630, 17], [645, 24], [660, 11],
[675, 22], [690, 17], [705, 12], [720, 10], [735, 13],
[750, 16], [765, 9], [780, 9], [795, 14], [810, 21],
[825, 17], [840, 13], [855, 12], [870, 18], [885, 10]])

# cost function with br8.data as the target
cost = decay.CostFactory2(data[:,0],data[:,1],5)


# End of file



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @uqfoundation)
# Copyright (c) 2020-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
misc user-defined items (solver configuration, moment constraints)
"""
from mystic.solvers import DifferentialEvolutionSolver2
from mystic.monitors import VerboseMonitor, Monitor
from mystic.termination import ChangeOverGeneration as COG

# kwds for solver
opts = dict(termination=COG(1e-6, 40), CrossProbability=0.9, ScalingFactor=0.9)
param = dict(solver=DifferentialEvolutionSolver2,
             npop=10, #XXX:npop
             maxiter=1000,
             maxfun=1e+6,
             x0=None, # use RandomInitialPoints
             nested=None, # don't use SetNested
             pool=None, # don't use SetMapper
             stepmon=VerboseMonitor(1, label='output'), # monitor config
             evalmon=Monitor(), # monitor config (re-initialized in solve)
             # kwds to pass directly to Solve(objective, **opt)
             opts=opts,
            )


from mystic.math.discrete import product_measure
from mystic.math import almostEqual as almost
from mystic.constraints import and_, integers
from mystic.coupler import outer

# lower and upper bound for parameters and weights
xlb = (20.0,0.0,2.1)
xub = (150.0,30.0,2.8)
wlb = (0,1,1)
wub = (1,1,1)
# number of Dirac masses to use for each parameter
npts = (2,1,1) #NOTE: rv = (w0,w0,x0,x0,w1,x1,w2,x2)
index = (5,)       #NOTE: rv[5] -> x1
# moments and uncertainty in first parameter
a_ave = None
a_var = None
a_ave_err = None
a_var_err = None
# moments and uncertainty in second parameter
b_ave = None
b_var = None
b_ave_err = None
b_var_err = None
# moments and uncertainty in output
o_ave = 6.5
o_ave_err = 1.0


def flatten(npts):
    'convert a moment constraint to a "flattened" constraint'
    def dec(f):
        def func(rv):
            c = product_measure().load(rv, npts)
            c = f(c)
            return c.flatten()
        return func
    return dec


def normalize_moments(mass=1.0, tol=1e-18, rel=1e-7):
    'normalize (using weights) on all measures'
    def func(c):
        for measure in c:
            if not almost(float(measure.mass), mass, tol=tol, rel=rel):
                measure.normalize()
        return c
    return func


def constrain_moments(ave=None, var=None, ave_err=None, var_err=None, idx=0):
    'impose mean and variance constraints on the selected measure'
    if ave is None: ave = float('nan')
    if var is None: var = float('nan')
    if ave_err is None: ave_err = 0
    if var_err is None: var_err = 0
    def func(c):
        E = float(c[idx].mean)
        if E > (ave + ave_err) or E < (ave - ave_err):
            c[idx].mean = ave
        E = float(c[idx].var)
        if E > (var + var_err) or E < (var - var_err):
            c[idx].var = var
        return c
    return func


#NOTE: model has single-value output; bounds = (lb,ub)
def constrain_expected(model, ave=None, ave_err=None, bounds=None):#, **kwds):
    'impose mean constraint on the expected value of the measure'
    if ave is None: ave = float('nan')
    if ave_err is None: ave_err = 0
    samples = None #kwds.get('samples', None) #NOTE: int or None
    if samples is None:
        def func(c):
            E = float(c.expect(model))
            if E > (ave + ave_err) or E < (ave - ave_err):
                c.set_expect(ave, model, bounds, tol=ave_err, npop=4)#XXX:npop
            return c
    else:
        return NotImplemented #FIXME: samples != None
        def func(c):
            E = float(c.sampled_expect(model, samples))
            if E > (ave + ave_err) or E < (ave - ave_err):
                c.set_sampled_expect(ave, model, bounds, tol=ave_err, npts=samples) #FIXME: NotImplemented
            return c
    return func


@integers(ints=float, index=index)
def integer_indices(rv):
    'constrain parameters at given index(es) to be ints'
    return rv


def constrained_integers(index=()):
    'check integer constraint is properly applied'
    def func(rv):
        return all(int(j) == j for i,j in enumerate(rv) if i in index)
    return func


def constrained(ave=None, var=None, ave_err=None, var_err=None, idx=0, debug=False):
    'check mean and variance on the selected measure are properly constrained'
    if ave is None: ave = float('nan')
    if var is None: var = float('nan')
    if ave_err is None: ave_err = 0
    if var_err is None: var_err = 0
    def func(c):
        E = float(c[idx].mean)
        if E > (ave + ave_err) or E < (ave - ave_err):
            if debug: print("skipping mean: %s" % E)
            return False
        E = float(c[idx].var)
        if E > (var + var_err) or E < (var - var_err):
            if debug: print("skipping var: %s" % E)
            return False
        return True
    return func


#NOTE: model has single-value output
def constrained_out(model, ave=None, ave_err=None, debug=False):#, **kwds):
    'check the expected output is properly constrained'
    if ave is None: ave = float('nan')
    if ave_err is None: ave_err = 0
    samples = None #kwds.get('samples', None) #NOTE: int or None
    if samples is None:
        def func(c):
            E = float(c.expect(model))
            if E > (ave + ave_err) or E < (ave - ave_err):
                if debug: print("skipping expected value: %s" % E)
                return False
            return True
    else:
        def func(c):
            E = float(c.sampled_expect(model, samples))
            if E > (ave + ave_err) or E < (ave - ave_err):
                if debug: print("skipping expected value: %s" % E)
                return False
            return True
    return func


def check(npts):
    'convert a moment check to a "flattened" check'
    def dec(f):
        def func(rv):
            c = product_measure().load(rv, npts)
            return f(c)
        return func
    return dec


# build a model representing 'truth'
from ouq_models import WrapModel
from surrogate import marc_surr as toy; nx = 3; ny = None
nargs = dict(nx=nx, ny=ny, rnd=False)
model = WrapModel('model', toy, **nargs)


# set the bounds
from mystic.bounds import MeasureBounds
bnd = MeasureBounds(xlb, xub, n=npts, wlb=wlb, wub=wub)


## moment-based constraints ##
normcon = normalize_moments()
###momcons = constrain_moments(a_ave, a_var, a_ave_err, a_var_err)
###is_cons = constrained(a_ave, a_var, a_ave_err, a_var_err)
#momcon0 = constrain_moments(a_ave, a_var, a_ave_err, a_var_err, idx=0)
#momcon1 = constrain_moments(b_ave, b_var, b_ave_err, b_var_err, idx=1)
#is_con0 = constrained(a_ave, a_var, a_ave_err, a_var_err, idx=0)
#is_con1 = constrained(b_ave, b_var, b_ave_err, b_var_err, idx=1)
#is_cons = lambda c: bool(additive(is_con0)(is_con1)(c))
momcons = constrain_expected(model, o_ave, o_ave_err, (bnd.lower,bnd.upper))
is_cons = constrained_out(model, o_ave, o_ave_err)

## index-based constraints ##
# impose constraints sequentially (faster, but assumes are decoupled)
#scons = outer(integer_indices)(flatten(npts)(outer(momcons)(normcon)))
scons = flatten(npts)(outer(momcons)(normcon))
#scons = flatten(npts)(outer(momcon1)(outer(momcon0)(normcon)))

# impose constraints concurrently (slower, but safer)
#ccons = and_(flatten(npts)(normcon), flatten(npts)(momcons), integer_indices)
ccons = and_(flatten(npts)(normcon), flatten(npts)(momcons))
#ccons = and_(flatten(npts)(normcon), flatten(npts)(momcon0), flatten(npts)(momcon1))

# check parameters (instead of measures)
iscon = check(npts)(is_cons)
#rvcon = constrained_integers(index)



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @uqfoundation)
# Copyright (c) 2020-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
OUQ classes for calculating bounds on statistical quantities
"""
from mystic.math.discrete import product_measure
from ouq import BaseOUQ


class MeanValue(BaseOUQ):
    def objective(self, rv, axis=None, idx=0): #FIXME: idx != None, is fixed
        """calculate mean value of input, under uncertainty

    Input:
        rv: list of input parameters
        axis: int, the index of output to calculate (all, by default)
        idx: int, the index of input to calculate (all, by default) #FIXME

    Returns:
        the mean value for the specified index

    NOTE:
        respects constraints on input parameters and product measure

    NOTE:
        for product_measure, use sampled_expect if samples, else expect
        """
        # check constraints
        c = product_measure().load(rv, self.npts)
        if not self.cvalid(c) or not self.xvalid(rv): #FIXME: set model,samples in constraints
            import numpy as np
            return np.inf
        # get mean value
        if axis is None and self.axes is not None:
            model = (lambda x: self.model(x, axis=i) for i in range(self.axes))
            if self.samples is None:
                return NotImplemented #FIXME: set model,samples in constraints
            # else use sampled support
            return NotImplemented #FIXME: set model,samples in constraints
        # else, get mean value for the given axis
        if axis is None:
            model = lambda x: self.model(x)
        else:
            model = lambda x: self.model(x, axis=axis)
        if self.samples is None:
            if idx is None:
                res = (c[i].mean for i in self.nx)
                return tuple(res)
            return c[idx].mean #FIXME: set model,samples in constraints
        return NotImplemented #FIXME: set model,samples in constraints




#!/usr/bin/env python
#
# Author: Patrick Hung (patrickh @caltech)
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
#
# Differential Evolution Strategies adapted from DESolver.py by Patrick Hung
"""
Differential Evolution Strategies

These strategies are to be passed into DifferentialEvolutionSolver's
Solve method, and determine how the candidate parameter values mutate
across a population.
"""

import random

def get_random_candidates(NP, exclude, N):
    """select N random candidates from population of size NP,
where exclude is the candidate to exclude from selection.

Thus, get_random_candidates(x,1,2) randomly selects two nPop[i],
where i != 1"""
    return random.sample(list(range(exclude))+list(range(exclude+1,NP)), N)


#################### #################### #################### ####################
#  Code below are the different crossovers/mutation strategies
#################### #################### #################### ####################

def Best1Exp(inst, candidate):
    """trial solution is current best solution plus scaled difference
of two randomly chosen candidates; mutates until random stop

trial = best + scale*(candidate1 - candidate2)"""
    r1,r2 = get_random_candidates(inst.nPop, candidate, 2) 
    n = random.randrange(inst.nDim)

    if inst._map_solver:
        trialSolution = inst.trialSolution[candidate]
    else:
        trialSolution = inst.trialSolution
    trialSolution[:] = inst.population[candidate]

    i = 0
    while 1:
        if random.random() >= inst.probability or i == inst.nDim:
            break
        trialSolution[n] = inst.bestSolution[n] + \
                           inst.scale * (inst.population[r1][n] - \
                                         inst.population[r2][n])
        n = (n + 1) % inst.nDim
        i += 1

#   inst._keepSolutionWithinRangeBoundary(inst.bestSolution)
    return

def Best1Bin(inst, candidate):
    """trial solution is current best solution plus scaled difference
of two randomly chosen candidates; mutates at random

trial = best + scale*(candidate1 - candidate2)"""
    # In DESolve, Best1Bin was identical to Best1Exp.
    # But the logic of Best1Bin is different from [1]. Reimplementing here.
    r1,r2 = get_random_candidates(inst.nPop, candidate, 2) 

    if inst._map_solver:
        trialSolution = inst.trialSolution[candidate]
    else:
        trialSolution = inst.trialSolution
    trialSolution[:] = inst.population[candidate]

    # Randomly chosen index between [0, ND-1] (See Eq.4 of [1] )
    n = random.randrange(inst.nDim)

    for i in range(inst.nDim):
        cross = random.random()
        if i==n or cross < inst.probability:
            # this component of trial vector will come from vector v
            trialSolution[i] = inst.bestSolution[i] + \
                               inst.scale * (inst.population[r1][i] - \
                                             inst.population[r2][i])

#   inst._keepSolutionWithinRangeBoundary(inst.bestSolution)
    return

def Rand1Exp(inst, candidate):
    """trial solution is randomly chosen candidate plus scaled difference
of two other randomly chosen candidates; mutates until random stop

trial = candidate1 + scale*(candidate2 - candidate3)"""
    r1,r2,r3 = get_random_candidates(inst.nPop, candidate, 3) 
    n = random.randrange(inst.nDim)

    if inst._map_solver:
        trialSolution = inst.trialSolution[candidate]
    else:
        trialSolution = inst.trialSolution
    trialSolution[:] = inst.population[candidate]

    i = 0
    while 1:
        if random.random() >= inst.probability or i == inst.nDim:
            break
        trialSolution[n] = inst.population[r1][n] + \
                           inst.scale * (inst.population[r2][n] - \
                                         inst.population[r3][n])
        n = (n + 1) % inst.nDim
        i += 1

#   inst._keepSolutionWithinRangeBoundary(inst.population[r1])
    return

# WARNING, stuff below are not debugged

def RandToBest1Exp(inst, candidate):
    """trial solution is itself plus scaled difference of best solution
and trial solution, plus the difference of two randomly chosen candidates;
mutates at random

trial += scale*(best - trial) + scale*(candidate1 - candidate2)"""
    r1,r2 = get_random_candidates(inst.nPop, candidate, 2) 
    n = random.randrange(inst.nDim)

    if inst._map_solver:
        trialSolution = inst.trialSolution[candidate]
    else:
        trialSolution = inst.trialSolution
    trialSolution[:] = inst.population[candidate][:]
    i = 0
    while 1:
        if random.random() >= inst.probability or i == inst.nDim:
            break
        trialSolution[n] += inst.scale * (inst.bestSolution[n] - \
                                          trialSolution[n]) + \
                            inst.scale * (inst.population[r1][n] - \
                                          inst.population[r2][n])
        n = (n + 1) % inst.nDim
        i += 1

#   inst._keepSolutionWithinRangeBoundary(inst.trialSolution)
    return

def Best2Exp(inst, candidate):
    """trial solution is current best solution plus scaled contributions
from four randomly chosen candidates; mutates until random stop

trial = best + scale*(candidate1 + candidate2 - candidate3 - candidate4)"""
    r1,r2,r3,r4 = get_random_candidates(inst.nPop, candidate, 4) 
    n = random.randrange(inst.nDim)

    if inst._map_solver:
        trialSolution = inst.trialSolution[candidate]
    else:
        trialSolution = inst.trialSolution
    trialSolution[:] = inst.population[candidate]
    i = 0
    while 1:
        if random.random() >= inst.probability or i == inst.nDim:
            break
        trialSolution[n] = inst.bestSolution[n] + \
                           inst.scale * (inst.population[r1][n] + \
                                         inst.population[r2][n] - \
                                         inst.population[r3][n] - \
                                         inst.population[r4][n])
        n = (n + 1) % inst.nDim
        i += 1

#   inst._keepSolutionWithinRangeBoundary(inst.bestSolution)
    return

def Rand2Exp(inst, candidate):
    """trial solution is randomly chosen candidate plus scaled contributions
from four other randomly chosen candidates; mutates until random stop

trial = candidate1 + scale*(candidate2 + candidate3 - candidate4 - candidate5)"""
    r1,r2,r3,r4,r5 = get_random_candidates(inst.nPop, candidate, 5) 
    n = random.randrange(inst.nDim)

    if inst._map_solver:
        trialSolution = inst.trialSolution[candidate]
    else:
        trialSolution = inst.trialSolution
    trialSolution[:] = inst.population[candidate]
    i = 0
    while 1:
        if random.random() >= inst.probability or i == inst.nDim:
            break
        trialSolution[n] = inst.population[r1][n] + \
                           inst.scale * (inst.population[r2][n] + \
                                         inst.population[r3][n] - \
                                         inst.population[r4][n] - \
                                         inst.population[r5][n])
        n = (n + 1) % inst.nDim
        i += 1

#   inst._keepSolutionWithinRangeBoundary(inst.population[r1])
    return

def Rand1Bin(inst, candidate):
    """trial solution is randomly chosen candidate plus scaled difference
of two other randomly chosen candidates; mutates at random

trial = candidate1 + scale*(candidate2 - candidate3)"""
    r1,r2,r3 = get_random_candidates(inst.nPop, candidate, 3) 
    n = random.randrange(inst.nDim)

    if inst._map_solver:
        trialSolution = inst.trialSolution[candidate]
    else:
        trialSolution = inst.trialSolution
    trialSolution[:] = inst.population[candidate]
    i = 0
    while 1:
        if random.random() >= inst.probability or i == inst.nDim:
            break
        trialSolution[n] = inst.population[r1][n] + \
                           inst.scale * (inst.population[r2][n] -\
                                         inst.population[r3][n])
        n = (n + 1) % inst.nDim
        i += 1

#   inst._keepSolutionWithinRangeBoundary(inst.population[r1])
    return

def RandToBest1Bin(inst, candidate):
    """trial solution is itself plus scaled difference of best solution
and trial solution, plus the difference of two randomly chosen candidates;
mutates until random stop

trial += scale*(best - trial) + scale*(candidate1 - candidate2)"""
    r1,r2 = get_random_candidates(inst.nPop, candidate, 2) 
    n = random.randrange(inst.nDim)

    if inst._map_solver:
        trialSolution = inst.trialSolution[candidate]
    else:
        trialSolution = inst.trialSolution
    trialSolution[:] = inst.population[candidate]
    i = 0
    while 1:
        if random.random() >= inst.probability or i == inst.nDim:
            break
        trialSolution[n] += inst.scale * (inst.bestSolution[n] - \
                                          trialSolution[n]) + \
                            inst.scale * (inst.population[r1][n] - \
                                          inst.population[r2][n])
        n = (n + 1) % inst.nDim
        i += 1

#   inst._keepSolutionWithinRangeBoundary(inst.trialSolution)
    return

def Best2Bin(inst, candidate):
    """trial solution is current best solution plus scaled contributions
of four randomly chosen candidates; mutates at random

trial = best + scale*(candidate1 - candidate2 - candidate3 - candidate4)"""
    r1,r2,r3,r4 = get_random_candidates(inst.nPop, candidate, 4) 
    n = random.randrange(inst.nDim)

    if inst._map_solver:
        trialSolution = inst.trialSolution[candidate]
    else:
        trialSolution = inst.trialSolution
    trialSolution[:] = inst.population[candidate]
    i = 0
    while 1:
        if random.random() >= inst.probability or i == inst.nDim:
            break
        trialSolution[n] = inst.bestSolution[n] + \
                           inst.scale * (inst.population[r1][n] + \
                                         inst.population[r2][n] - \
                                         inst.population[r3][n] - \
                                         inst.population[r4][n])
        n = (n + 1) % inst.nDim
        i += 1

#   inst._keepSolutionWithinRangeBoundary(inst.bestSolution)
    return

def Rand2Bin(inst, candidate):
    """trial solution is randomly chosen candidate plus scaled contributions
of four other randomly chosen candidates; mutates at random

trial = candidate1 + scale*(candidate2 - candidate3 - candidate4 - candidate5)"""
    r1,r2,r3,r4,r5 = get_random_candidates(inst.nPop, candidate, 5) 
    n = random.randrange(inst.nDim)

    if inst._map_solver:
        trialSolution = inst.trialSolution[candidate]
    else:
        trialSolution = inst.trialSolution
    trialSolution[:] = inst.population[candidate]
    i = 0
    while 1:
        if random.random() >= inst.probability or i == inst.nDim:
            break
        trialSolution[n] = inst.population[r1][n] + \
                           inst.scale * (inst.population[r2][n] + \
                                         inst.population[r3][n] - \
                                         inst.population[r4][n] - \
                                         inst.population[r5][n])
        n = (n + 1) % inst.nDim
        i += 1

#   inst._keepSolutionWithinRangeBoundary(inst.population[r1])
    return
    
# end of file



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @uqfoundation)
# Copyright (c) 2020-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
cartesian bounds and measure bounds instances
"""
__all__ = ['Bounds','MeasureBounds']

from mystic.tools import flatten


class Bounds(object):

    def __init__(self, *bounds, **kwds):
        """create a bounds instance

    bounds is a tuple of (lower, upper) bound

    additionally, one can specify:
        - xlb: lower bound
        - xub: upper bound
        - n: repeat

    For example:
        >>> b = Bounds(7, 8, n=2)
        >>> b.xlb, b.xub
        (7, 8)
        >>> b()
        [(7, 8), (7, 8)]
        >>> b.lower
        [7, 7]
        >>> b.upper
        [8, 8]
        >>> 
        >>> c = Bounds((0,1),(3,4), n=2)
        >>> c()
        [(0, 3), (0, 3), (1, 4), (1, 4)]
        """
        self.n = kwds.pop('n', 1)
        self.wub = None # 1.0
        self.wlb = None # 1.0

        xlb = -float('inf')
        xub = float('inf')

        blen = len(bounds)
        if blen == 0:
            self.xub = kwds.pop('xub', xub)
            self.xlb = kwds.pop('xlb', xlb)
        elif blen == 1:
            self.xlb = bounds[0]
            self.xub = kwds.pop('xub', xub)
            if 'xlb' in kwds:
                msg = "__init__() got multiple values for argument 'xlb'"
                raise TypeError(msg)
        else:
            self.xlb,self.xub = bounds
            if 'xlb' in kwds:
                msg = "__init__() got multiple values for argument 'xlb'"
                raise TypeError(msg)
            if 'xub' in kwds:
                msg = "__init__() got multiple values for argument 'xub'"
                raise TypeError(msg)
        # single xub stretched to xlb
        if hasattr(self.xlb, '__len__') and not hasattr(self.xub, '__len__'):
            self.xub = len(self.xlb) * (self.xub,)
        # single xlb stretched to xub
        elif hasattr(self.xub, '__len__') and not hasattr(self.xlb, '__len__'):
            self.xlb = len(self.xub) * (self.xlb,)
        # single n stretched to xub
        if hasattr(self.xub, '__len__') and not hasattr(self.n, '__len__'):
            self.n = len(self.xub) * (self.n,)
        # single xlb,xub stretched to n
        elif hasattr(self.n, '__len__') and not hasattr(self.xub, '__len__'):
            self.xub = len(self.n) * (self.xub,)
            self.xlb = len(self.n) * (self.xlb,)

    def __lower(self):
        "get list of lower bounds"
        n = (self.n,) if not hasattr(self.n, '__len__') else self.n
        xlb = (self.xlb,) if not hasattr(self.xlb, '__len__') else self.xlb
        return list(flatten(i*[j] for i,j in zip(n,xlb)))

    def __upper(self):
        "get list of upper bounds"
        n = (self.n,) if not hasattr(self.n, '__len__') else self.n
        xub = (self.xub,) if not hasattr(self.xub, '__len__') else self.xub
        return list(flatten(i*[j] for i,j in zip(n,xub)))

    def __call__(self):
        "get list of tuples of (lower, upper) bounds"
        return list(zip(self.lower, self.upper))

    def __add__(self, other): #FIXME: create new Bounds instance
        if not isinstance(other, Bounds):
            return NotImplemented
        return self() + other()

    def __set_lower(self, lb):
        return NotImplemented

    def __set_upper(self, ub):
        return NotImplemented

    def __repr__(self):
        return "%s(%s, %s, n=%s)" % (self.__class__.__name__, self.xlb, self.xub, self.n)

    lower = property(__lower, __set_lower)
    upper = property(__upper, __set_upper)


class MeasureBounds(Bounds):

    def __init__(self, *bounds, **kwds):
        """create a measure bounds instance

    bounds is a tuple of (lower, upper) bound

    additionally, one can specify:
        - wlb: weight lower bound
        - wub: weight upper bound
        - xlb: lower bound
        - xub: upper bound
        - n: repeat

    For example:
        >>> b = MeasureBounds(7, 8, n=2)
        >>> b.wlb, b.wub
        (0, 1)
        >>> b.xlb, b.xub
        (7, 8)
        >>> b()
        [(0, 1), (0, 1), (7, 8), (7, 8)]
        >>> b.lower
        [0, 0, 7, 7]
        >>> b.upper
        [1, 1, 8, 8]
        >>> 
        >>> c = MeasureBounds((0,1),(4,5), n=1, wlb=(0,1), wub=(2,3))
        >>> c.lower
        [0, 0, 1, 1]
        >>> c.upper
        [2, 4, 3, 5]
        >>> c()
        [(0, 2), (0, 4), (1, 3), (1, 5)]
        >>> 
        >>> c = MeasureBounds((0,1),(4,5), n=2)
        >>> c()
        [(0, 1), (0, 1), (0, 4), (0, 4), (0, 1), (0, 1), (1, 5), (1, 5)]
        """
        super(MeasureBounds, self).__init__(*bounds, **kwds)
        self.wub = kwds.pop('wub', 1)
        self.wlb = kwds.pop('wlb', 0)
        # single wlb stretched to wub
        if hasattr(self.wub, '__len__') and not hasattr(self.wlb, '__len__'):
            self.wlb = len(self.wub) * (self.wlb,)
        elif hasattr(self.wlb, '__len__') and not hasattr(self.wub, '__len__'):
            self.wub = len(self.wlb) * (self.wub,)
        # single wlb stretched to xlb
        if hasattr(self.xlb, '__len__') and not hasattr(self.wlb, '__len__'):
            self.wlb = len(self.xlb) * (self.wlb,)
            self.wub = len(self.xub) * (self.wub,)
        elif hasattr(self.wlb, '__len__') and not hasattr(self.xlb, '__len__'):
            self.n = len(self.wlb) * (self.n,)
            self.xlb = len(self.wlb) * (self.xlb,)
            self.xub = len(self.wub) * (self.xub,)

    def __lower(self):
        n = (self.n,) if not hasattr(self.n, '__len__') else self.n
        wlb = (self.wlb,) if not hasattr(self.wlb, '__len__') else self.wlb
        xlb = (self.xlb,) if not hasattr(self.xlb, '__len__') else self.xlb
        return list(flatten(i*[j] + i*[k] for i,j,k in zip(n,wlb,xlb)))

    def __upper(self):
        n = (self.n,) if not hasattr(self.n, '__len__') else self.n
        wub = (self.wub,) if not hasattr(self.wub, '__len__') else self.wub
        xub = (self.xub,) if not hasattr(self.xub, '__len__') else self.xub
        return list(flatten(i*[j] + i*[k] for i,j,k in zip(n,wub,xub)))

    def __add__(self, other): #FIXME: create new MeasureBounds instance
        if not isinstance(other, Bounds):
            return NotImplemented
        return self() + other()

    def __set_lower(self, lb):
        return NotImplemented

    def __set_upper(self, ub):
        return NotImplemented

    def __repr__(self):
        if self.wlb == 0 and self.wub == 1:
            return super(MeasureBounds, self).__repr__()
        return "%s(%s, %s, n=%s, wlb=%s, wub=%s)" % (self.__class__.__name__, self.xlb, self.xub, self.n, self.wlb, self.wub)

    lower = property(__lower, __set_lower)
    upper = property(__upper, __set_upper)




#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Author: Alta Fang (altafang @caltech and alta @princeton)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
shortcut (?) math tools related to statistics;
also, math tools related to gaussian distributions
"""
import math

#----------------------------------------------------------------
# Gaussian helper functions

def erf(x):
    "evaluate the error function at x" 
    try:                                               
        from math import erf
    except ImportError:
        try:
            from ctypes import util, CDLL, c_double
            erf = CDLL(util.find_library('m')).erf
            erf.argtypes = [c_double]
            erf.restype = c_double
        except ImportError:
            erf = _erf
    if hasattr(x, '__len__'):
        from numpy import vectorize
        erf = vectorize(erf)
    return erf(x)

def gamma(x):
    "evaluate the gamma function at x"
    try:
        from math import gamma
    except ImportError:
        try:
            from ctypes import util, CDLL, c_double
            gamma = CDLL(util.find_library('m')).gamma
            gamma.argtypes = [c_double]
            gamma.restype = c_double
        except ImportError:
            gamma = _gamma
    if hasattr(x, '__len__'):
        from numpy import vectorize
        gamma = vectorize(gamma)
    return gamma(x) 

def lgamma(x):
    "evaluate the natual log of the abs value of the gamma function at x"
    try:
        from math import lgamma
    except ImportError:
        try:
            from ctypes import util, CDLL, c_double
            lgamma = CDLL(util.find_library('m')).lgamma
            lgamma.argtypes = [c_double]
            lgamma.restype = c_double
        except ImportError:
            lgamma = _lgamma
    if hasattr(x, '__len__'):
        from numpy import vectorize
        lgamma = vectorize(lgamma)
    return lgamma(x) 


def _erf(x):
    "approximate the error function at x"
    # Source: http://picomath.org/python/erf.py.html
    a1 =  0.254829592
    a2 = -0.284496736
    a3 =  1.421413741
    a4 = -1.453152027
    a5 =  1.061405429
    p  =  0.3275911

    # Save the sign of x
    sign = 1
    if x < 0:
        sign = -1
    x = abs(x)

    # A&S formula 7.1.26
    t = 1.0/(1.0 + p*x)
    y = 1.0 - (((((a5*t + a4)*t) + a3)*t + a2)*t + a1)*t*math.exp(-x*x)
    return sign*y


def _gamma(x):
    "approximate the gamma function at x"
    # Source: http://picomath.org/python/gamma.py.html
    if x <= 0:
        raise ValueError("Invalid input")

    # Split the function domain into three intervals:
    # (0, 0.001), [0.001, 12), and (12, infinity)

    ###########################################################################
    # First interval: (0, 0.001)
    #
    # For small x, 1/Gamma(x) has power series x + gamma x^2  - ...
    # So in this range, 1/Gamma(x) = x + gamma x^2 with error order of x^3.
    # The relative error over this interval is less than 6e-7.

    gamma = 0.577215664901532860606512090 # Euler's gamma constant

    if x < 0.001:
        return 1.0/(x*(1.0 + gamma*x))

    ###########################################################################
    # Second interval: [0.001, 12)

    if x < 12.0:
        # The algorithm directly approximates gamma over (1,2) and uses
        # reduction identities to reduce other arguments to this interval.
        
        y = x
        n = 0
        arg_was_less_than_one = (y < 1.0)

        # Add or subtract integers as necessary to bring y into (1,2)
        # Will correct for this below
        if arg_was_less_than_one:
            y += 1.0
        else:
            n = int(math.floor(y)) - 1  # will use n later
            y -= n

        # numerator coefficients for approximation over the interval (1,2)
        p = [
            -1.71618513886549492533811E+0,
             2.47656508055759199108314E+1,
            -3.79804256470945635097577E+2,
             6.29331155312818442661052E+2,
             8.66966202790413211295064E+2,
            -3.14512729688483675254357E+4,
            -3.61444134186911729807069E+4,
             6.64561438202405440627855E+4
        ]

        # denominator coefficients for approximation over the interval (1,2)
        q = [
            -3.08402300119738975254353E+1,
             3.15350626979604161529144E+2,
            -1.01515636749021914166146E+3,
            -3.10777167157231109440444E+3,
             2.25381184209801510330112E+4,
             4.75584627752788110767815E+3,
            -1.34659959864969306392456E+5,
            -1.15132259675553483497211E+5
        ]

        num = 0.0
        den = 1.0

        z = y - 1
        for i in range(8):
            num = (num + p[i])*z
            den = den*z + q[i]
        result = num/den + 1.0

        # Apply correction if argument was not initially in (1,2)
        if arg_was_less_than_one:
            # Use identity gamma(z) = gamma(z+1)/z
            # The variable "result" now holds gamma of the original y + 1
            # Thus we use y-1 to get back the orginal y.
            result /= (y-1.0)
        else:
            # Use the identity gamma(z+n) = z*(z+1)* ... *(z+n-1)*gamma(z)
            for _ in range(n):
                result *= y
                y += 1

        return result

    ###########################################################################
    # Third interval: [12, infinity)

    if x > 171.624:
        # Correct answer too large to display. 
        return 1.0/0 # float infinity

    return math.exp(log_gamma(x))


def _lgamma(x):
    "approximate the natual log of the abs value of the gamma function at x"
    # Source: http://picomath.org/python/gamma.py.html
    if x <= 0:
        raise ValueError("Invalid input")

    if x < 12.0:
        return math.log(abs(gamma(x)))

    # Abramowitz and Stegun 6.1.41
    # Asymptotic series should be good to at least 11 or 12 figures
    # For error analysis, see Whittiker and Watson
    # A Course in Modern Analysis (1927), page 252

    c = [
         1.0/12.0,
        -1.0/360.0,
         1.0/1260.0,
        -1.0/1680.0,
         1.0/1188.0,
        -691.0/360360.0,
         1.0/156.0,
        -3617.0/122400.0
    ]
    z = 1.0/(x*x)
    sum = c[7]
    for i in range(6, -1, -1):
        sum *= z
        sum += c[i]
    series = sum/x

    halfLogTwoPi = 0.91893853320467274178032973640562
    logGamma = (x - 0.5)*math.log(x) - x + halfLogTwoPi + series
    return logGamma

#----------------------------------------------------------------

def _lefttail(percent):
    "calculate left-area percent from center-area percent"
    return 100 - .5*(100 - percent)

def stderr(std, npts):
    "standard error"
    return std/npts**.5

def meanconf(std, npts, percent=95):
    "mean confidence interval: returns conf, where interval = mean +/- conf"
    import scipy.stats as ss
    scale = ss.norm.ppf(.01*_lefttail(percent))
    return scale * stderr(std, npts)

def sampvar(var, npts):
    "sample variance from variance"
    return var * npts/(npts-1)

def _varconf(var, npts, percent=95):
    "var confidence interval: returns interval, where var in interval"
    s = var * npts
    import scipy.stats as ss
    chihi = ss.chi2.isf(.01*_lefttail(percent), npts-1)
    chilo = ss.chi2.isf(1 - .01*_lefttail(percent), npts-1)
    return s/chihi, s/chilo

def varconf(var, npts, percent=95, tight=False):
    "var confidence interval: returns max interval distance from var"
    hi,lo = _varconf(var, npts, percent)
    select = min if tight else max
    return select(abs(hi-var),abs(var-lo))

#----------------------------------------------------------------

def cdf_factory(mean, variance):
  """
Returns cumulative distribution function (as a Python function)
for a Gaussian, given the mean and variance
"""
  def cdf(x):
    """Cumulative distribution function."""
    try:
      from scipy.special import erf
    except ImportError:
      pass
    return 0.5*(1. + erf((x-mean)/math.sqrt(2.*variance)))
  return cdf


def pdf_factory(mean, variance):
  """
Returns a probability density function (as a Python function)
for a Gaussian, given the mean and variance
"""
  def pdf(x):
    """Probability density function."""
    return 1./math.sqrt(2.*math.pi*variance)* \
           math.exp(-(x - mean)**2/(2.*variance))
  return pdf


#----------------------------------------------------------------
# shortcut (?) statistics functions

def volume(lb,ub):
  """calculates volume for a uniform distribution in n-dimensions"""
  vol = 1
  for i in range(len(ub)):
    vol *= abs(ub[i] - lb[i])
  return vol


def prob_mass(volume,norm):
  """calculates probability mass given volume and norm"""
  return volume / norm  #XXX: norm != 0


def mean(expectation,volume):
  """calculates mean given expectation and volume"""
  return expectation / volume  #XXX: volume != 0


def mcdiarmid_bound(mean,diameter):
  """calculates McDiarmid bound given mean and McDiarmid diameter"""
  if not diameter: return 1.0  #XXX: define e^(0/0) = 1
  return math.exp(-2.0 * (max(0,mean))**2 / diameter**2) #XXX: or diameter**1 ?


def __mean(xarr):
  """mean = x.sum() / len(x)""" #interface is [lb,ub]; not lb,ub
  from numpy import mean
  return mean(xarr)


def __variance(xarr):
  """var = mean(abs(x - x.mean())**2) / 3""" #interface is [lb,ub]; not lb,ub
  from numpy import var
  return var(xarr) / 3.0


#----------------------------------------------------------------------------
# Tests

def __test_probability_mass():
  bounds = [(0.1,1.0),(2.0,4.0),(4.0,9.0)]
  lower = [i[0] for i in bounds]
  upper = [i[1] for i in bounds]

  cuboid_volume = volume(lower,upper)
  probability_mass = prob_mass(cuboid_volume,cuboid_volume)
  print("probability mass: %s" % probability_mass)


if __name__ == '__main__':
  __test_probability_mass()

# EOF



#!/usr/bin/env python
#
# Author: Patrick Hung (patrickh @caltech)
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Same as test_lorentzian, but with n being a fitted variable

This is MUCH faster than test_lorentzian because the cost function no
longer has to do an "integral" as an intermediate step
"""

import matplotlib.pyplot as plt, matplotlib
from numpy import *
from mystic.solvers import DifferentialEvolutionSolver
from mystic.termination import ChangeOverGeneration, VTR
from mystic.strategy import Best1Exp, Rand1Exp, Best2Exp, Best2Exp
from mystic.monitors import Monitor
from mystic.tools import getch

from mystic.tools import random_seed
random_seed(123)

#matplotlib.interactive(True)

from mystic.models import lorentzian
F = lorentzian.ForwardFactory

def show():
    import Image
    plt.savefig('test_lorentzian_out',dpi=72)
    im = Image.open('test_lorentzian_out.png')
    im.show()
    return

def plot_sol(solver=None):
    def _(params):
        import mystic._signal as signal
        print("plotting params: %s" % params)
        plt.errorbar(binsc, histo, sqrt(histo), fmt='b+')
        x = arange(xmin, xmax, (0.1* binwidth))
        plt.plot(x, pdf(x)*N,'b:')
        plt.plot(x, F(params)(x)*N,'r-')
        plt.xlabel('E (GeV)')
        plt.ylabel('Counts')
        try: show()
        except ImportError: plt.show()
        if solver is not None:
            signal.signal(signal.SIGINT, signal.Handler(solver))
    return _

ND = 7
NP = 50
MAX_GENERATIONS = 200
generations = 200
minrange = [0.5,30, -15, 10, 0, 0, 100]
maxrange = [2, 60., -5, 50, 2, 1, 200]

def de_solve(CF):
    solver = DifferentialEvolutionSolver(ND, NP)
    solver.enable_signal_handler()
    stepmon = Monitor()
    solver.SetRandomInitialPoints(min=minrange,max=maxrange)
    solver.SetStrictRanges(min=minrange,max=maxrange)
    solver.SetEvaluationLimits(generations=MAX_GENERATIONS)
    solver.SetGenerationMonitor(stepmon)
    termination=ChangeOverGeneration(generations=generations)
    solver.Solve(CF, termination=termination, strategy=Rand1Exp, \
                 sigint_callback = plot_sol(solver))
    solution = solver.Solution()
    return solution, stepmon


if __name__ == '__main__':
    target = [1., 45., -10., 20., 1., 0.1, 120.]

    from mystic.models.lorentzian import gendata, histogram
    npts = 4000; binwidth = 0.1
    N = npts * binwidth
    xmin, xmax = 0.0, 3.0
    pdf = F(target)
    print("pdf(1): %s" % pdf(1))

    data = gendata(target, xmin, xmax, npts)
    plt.plot(data[1:int(N)],0*data[1:int(N)],'k.')
    plt.title('Samples drawn from density to be estimated.')
    try: show()
    except ImportError: plt.show()
    plt.clf()

    binsc, histo = histogram(data, binwidth, xmin,xmax)
    print("binsc:  %s" % binsc)
    print("count:  %s" % histo)
    print("ncount: %s" % (histo//N))
    print("exact : %s" % pdf(binsc))

    print("now with DE...")
    myCF = lorentzian.CostFactory2(binsc, histo//N, ND)
    sol, steps = de_solve(myCF)
    plot_sol()(sol)
    #print("steps: %s %s" % (steps.x, steps.y))

# end of file



#!/usr/bin/env python
#
# Author: Patrick Hung (patrickh @caltech)
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Sets up Zimmermann's problem. This is problem 8 of testbed 1 in [1] and [2].

Solution: Min of 0 @ Vector[0]

Reference:

[1] Storn, R. and Price, K. Differential Evolution - A Simple and Efficient
Heuristic for Global Optimization over Continuous Spaces. Journal of Global
Optimization 11: 341-359, 1997.

[2] Storn, R. and Proce, K. Same title as above, but as a technical report.
try: http://www.icsi.berkeley.edu/~storn/deshort1.ps
"""

from mystic.solvers import DifferentialEvolutionSolver
from mystic.termination import ChangeOverGeneration, VTR
from mystic.strategy import Best1Exp, Rand1Exp

from mystic.tools import random_seed
random_seed(123)

# Eq. (24-26) of [2].
from mystic.models import zimmermann as CostFunction

ND = 2
NP = 20
MAX_GENERATIONS = 2500

def main():
    solver = DifferentialEvolutionSolver(ND, NP)

    solver.SetRandomInitialPoints(min = [0.]*ND, max = [5.]*ND)
    solver.SetEvaluationLimits(generations=MAX_GENERATIONS)

    solver.Solve(CostFunction, termination=VTR(0.0000001), strategy=Rand1Exp, \
                 CrossProbability=0.3, ScalingFactor=1.0)

    solution = solver.Solution()
  
    print(solution)



if __name__ == '__main__':
    from timeit import Timer
    t = Timer("main()", "from __main__ import main")
    timetaken =  t.timeit(number=1)
    print("CPU Time: %s" % timetaken)

    from mystic.monitors import Monitor
    from mystic.solvers import NelderMeadSimplexSolver as fmin
    from mystic.termination import CandidateRelativeTolerance as CRT

    import random
    simplex = Monitor()
    esow = Monitor()
    xinit = [random.uniform(0,5) for j in range(ND)]

    solver = fmin(len(xinit))
    solver.SetInitialPoints(xinit)
    solver.SetEvaluationMonitor(esow)
    solver.SetGenerationMonitor(simplex)
    solver.Solve(CostFunction, CRT())
    sol = solver.Solution()
    print("fmin solution: %s" % sol)

# end of file



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @uqfoundation)
# Copyright (c) 2020-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
'''
hyperparameter tuning for least upper bound of ExpectedValue on model

Test function is y = F(x), where:
  y0 = x0 + x1 * | x2 * x3**2 - (x4 / x1)**2 |**.5
  y1 = x0 - x1 * | x2 * x3**2 + (x4 / x1)**2 |**.5
  y2 = x0 - | x1 * x2 * x3 - x4 |

toy = lambda x: F(x)[0]
model = lambda x: toy(x), with x[-2:] = (d,e)
for hyperparameters z = [d, e]

Calculate least upper bound on E|model(x)|, where:
  x in [(0,1), (1,10), (0,10)]
  wx in [(0,1), (1,1), (1,1)]
  npts = [2, 1, 1] (i.e. two Dirac masses on x[0], one elsewhere)
  sum(wx[i]_j) for j in [0,npts], for each i
  mean(x[0]) = 5e-1 +/- 1e-3
  var(x[0]) = 5e-3 +/- 1e-4
  z in [(0,10), (0,10)]

Solves for z that produces least upper bound on E|model(x|z)|,
given the bounds, normalization, and moment constraints.

Creates 'log' of inner optimizations, and 'result' for outer optimization.

Check results while running (using log reader):
  $ mystic_log_reader log.txt -n 0 -g -p '0:2,2:4,5,7'
  $ mystic_log_reader result.txt -g -p '0,1'
'''
from ouq_models import *


if __name__ == '__main__':

    #from toys import cost5x3 as toy; nx = 5; ny = 3
    #from toys import function5x3 as toy; nx = 5; ny = 3
    #from toys import cost5x1 as toy; nx = 5; ny = 1
    #from toys import function5x1 as toy; nx = 5; ny = 1
    #from toys import cost5 as toy; nx = 5; ny = None
    from toys import function5 as toy; nx = 5; ny = None
    from toys import wrap; nx = nx-2 #NOTE: will reduce nx by 2

    # update 'inner-loop' optimization parameters
    from misc import param, npts, wlb, wub, is_cons, scons
    from ouq import ExpectedValue
    from mystic.bounds import MeasureBounds
    from mystic.monitors import VerboseLoggingMonitor, Monitor, VerboseMonitor
    from mystic.termination import VTRChangeOverGeneration as VTRCOG
    from mystic.termination import Or, VTR, ChangeOverGeneration as COG
    param['opts']['termination'] = COG(1e-10, 100) #NOTE: short stop?
    param['npop'] = 40 #NOTE: increase if results.txt is not monotonic
    param['stepmon'] = VerboseLoggingMonitor(1, 20, filename='log.txt', label='output')

    # build inner-loop and outer-loop bounds
    bnd = MeasureBounds((0,1,0,0,0)[:nx],(1,10,10,10,10)[:nx], n=npts[:nx], wlb=wlb[:nx], wub=wub[:nx])
    bounds = [(0,10),(0,10)] #NOTE: x[-2],x[-1]

    # get initial guess, a monitor, and a counter
    import mystic._counter as it
    counter = it.Counter()
    import numpy as np
    in_bounds = lambda a,b: (b-a) * np.random.rand() + a
    from pathos.pools import ProcessPool as Pool
    from mystic.solvers import fmin_powell, lattice, PowellDirectionalSolver
    axis = 0 #FIXME: calculation axis (for upper_bound, and thus cost)
    Ns = 25 #XXX: number of samples, when model has randomness
    #_solver, x0 = diffev2, bounds
    #_solver, x0 = fmin_powell, [.5 * sum(i) for i in bounds]
    _solver, x0 = fmin_powell, [in_bounds(*i) for i in bounds]
    #_solver, x0 = lattice, len(bounds)
    stepmon = VerboseLoggingMonitor(1, 1, 1, filename='result.txt', label='solved')

    def cost(x, axis=None, samples=Ns):
        """upper bound on expected model output

    Inputs:
        x: list of model hyperparameters
        axis: int, the index of y on which to find bound (all, by default)
        samples: int, number of samples, for a non-deterministic OUQ model

    Returns:
        upper bound on expected value of model output
        """
        # build a model, F(x|a), and tune "a" for optimal F.
        toy_ = wrap(d=x[0], e=x[1])(toy) #NOTE: reduces nx by 2
        #print('building model F(x|a)...')
        model = WrapModel('model', model=toy_, nx=nx, ny=ny, rnd=False)

        rnd = Ns if model.rnd else None
        #print('building UQ objective of expected model output...')
        b = ExpectedValue(model, bnd, constraint=scons, cvalid=is_cons, samples=rnd)
        i = counter.count()
        #print('solving for upper bound on expected model output...')
        solved = b.upper_bound(axis=axis, id=i, **param)
        if type(solved) is not tuple:
            solved = (solved,)
        if axis is None:
            results = tuple(-s for s in solved) #NOTE: -1 for LUB
        else:
            results = -solved[axis] #NOTE: -1 for LUB
        return results

    # outer-loop solver configuration and execution
    settings = dict(args=(axis,Ns), bounds=bounds, maxiter=1000, maxfun=100000,
                    disp=1, full_output=1, itermon=stepmon, ftol=1e-6,
                    npop=4, gtol=4, solver=PowellDirectionalSolver)# xtol=1e-6)
    pool = Pool(settings['npop'])
    _map = pool.map
    result = _solver(cost, x0, map=_map, **settings)
    pool.close(); pool.join(); pool.clear()

    # get the best result (generally the same as returned by solver)
    m = stepmon.min()
    print("%s @ %s" % (m.y, m.x)) #NOTE: -1 for max, 1 for min
    #print("%s @ %s" % (result[1], result[0])) #NOTE: -1 for max, 1 for min




#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @uqfoundation)
# Copyright (c) 2020-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
'''
calculate error for actively learned/interpolated models

Test function is y = F(x), where:
  y0 = x0 + x1 * | x2 * x3**2 - (x4 / x1)**2 |**.5
  y1 = x0 - x1 * | x2 * x3**2 + (x4 / x1)**2 |**.5
  y2 = x0 - | x1 * x2 * x3 - x4 |

toy = lambda x: F(x)[0]
golden = lambda x: toy(x + .001) - .001
truth = lambda x: G(toy(G(x + .001, .01)) -.001, .01)
G(mu, sigma) is a Gaussian with mean = mu and std = sigma

1) Sample 10 pts in [0,10] with truth.
   Find graphical distance between truth and sampled pts.
2) Sample golden with 4 solvers, then interpolate to produce a surrogate.
   Find pointwise distance (golden(x) - surrogate(x))**2.
3) Sample golden with 4 more solvers, then interpolate an updated surrogate.
   Find pointwise distance (golden(x) - surrogate(x))**2.
4) Train a MLP Regressor on the sampled data.
   Find pointwise distance (golden(x) - surrogate(x))**2.

Creates 'golden' and 'truth' databases of stored evaluations.
'''
from ouq_models import *


if __name__ == '__main__':

    #from toys import cost5x3 as toy; nx = 5; ny = 3
    #from toys import function5x3 as toy; nx = 5; ny = 3
    #from toys import cost5x1 as toy; nx = 5; ny = 1
    #from toys import function5x1 as toy; nx = 5; ny = 1
    #from toys import cost5 as toy; nx = 5; ny = None
    from toys import function5 as toy; nx = 5; ny = None

    # build a model representing 'truth' (one deterministic, and one not)
    truth = dict(model=toy, nx=nx, ny=ny, mu=.001, zmu=-.001)#, uid=True)
    golden = NoisyModel('golden', cached=True, sigma=0, zsigma=0, **truth)
    truth = NoisyModel('truth', sigma=.01, zsigma=.01, **truth)

    # generate data (DB) of sampled 'truth'
    deterministic = False
    Gx = golden if deterministic else truth
    bounds = [(0,10)]*nx
    data = Gx.sample(bounds, pts=10) #FIXME: activate cache w/o calling sample?
    print("size of data: %s" % len(data.coords))
    #print(len(data.values))

    # get graphical distance (for 'truth')
    error = Gx.distance(data, axis=None)
    import numpy as np
    print('total error: %s' % np.sum(error))
    print('max error: %s' % np.max(error, axis=-1))

    # calculate model error for 'golden'
    data = golden.sample(bounds, pts=-4)
    #print('truth: %s' % str(golden([1,2,3,4,5])))
    estimate = dict(nx=nx, ny=ny, data=golden, noise=0, smooth=0)
    surrogate = InterpModel('surrogate', method='thin_plate', **estimate)
    print('estimate: %s' % str(surrogate([1,2,3,4,5])))
    print('truth: %s' % str(golden([1,2,3,4,5])))
    error = dict(model=golden, surrogate=surrogate)
    misfit = ErrorModel('misfit', **error)
    print('error: %s' % str(misfit([1,2,3,4,5])))

    #'''
    # sample more data, refit, and recalculate error
    print('resampling and refitting surrogate')
    data = golden.sample(bounds, pts=-4)
    surrogate.fit()
    print('estimate: %s' % str(surrogate([1,2,3,4,5])))
    print('error: %s' % str(misfit([1,2,3,4,5])))
    #'''

    #'''
    # fit a learned model using the sampled data, and calculate error
    print('fitting an estimator with machine learning')
    estimate = dict(nx=nx, ny=ny, data=golden)
    mlarg = dict(hidden_layer_sizes=(100,75,50,25),  max_iter=1000, n_iter_no_change=5, solver='lbfgs')
    import sklearn.neural_network as nn
    estimator = nn.MLPRegressor(**mlarg)
    learned = LearnedModel('learned', estimator=estimator, **estimate)
    print('estimate: %s' % str(learned([1,2,3,4,5])))
    mlerror = dict(model=golden, surrogate=learned)
    error = ErrorModel('error', **mlerror)
    print('error: %s' % str(error([1,2,3,4,5])))
    #'''




#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2010-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
an interpolator
  - initalize with objective f(x) (and 'Sampler' object)
  - can attach a monitor and/or archiver
  - can sample points (using the Sampler)
  - can downsample and/or add noise
  - interpolates with "interp.interp"
  - converts f(*x) <-> f(x)
  - plot data and interpolated surface
"""

class Surface(object): #FIXME: should be subclass of Interpolator (?)
   #surface has:
   #    args - interpolation configuration (smooth, function, ...)
   #    sampler - a search algorithm
   #    maxpts - a maximum number of sampling points
   #    noise - a noise coefficient
   #    dim - dimensionality of the model
   #    function - target function [F(*x)]
   #    objective - target function [F(x)]
   #    surrogate - interpolated function [F(*x)]
   #    model - interpolated function [F(x)]
   #    x,y,z - sampled points(*)
   #
   #surface (or sampler) has:
   #    _minmon,_maxmon - step monitor(*)
   #    _minarch,_maxarch - sampled point archives(*)
   #
   #surface can:
   #    Sample - populate sampled point archive with solver trajectories
   #    Interpolate - build interpolated function from sampled points
   #    Plot - plot sampled points and interpolated surface
   #
   #surface (or sampler) can:
   #    UseMonitor - track trajectories with a monitor(s)
   #    UseArchive - track sampled points in an archive(s)
   #    _max - fetch (x,y,z,model(x,y)) for maximal z of sampled points
   #    _min - fetch (x,y,z,model(x,y)) for minimal z of sampled points

    def __init__(self, objective, sampler=None, **kwds):
        """response surface interpolator, where data is sampled from objective

        Input:
          objective: function of the form z=f(x)
          sampler: mystic.search.Searcher instance

        Additional Inputs:
          maxpts: int, maximum number of points to use from (x,z)
          noise: float, amplitude of gaussian noise to remove duplicate x
          method: string for kind of interpolator
          dim: number of parameters in the input for the objective function
          filter: a data filter produced with mystic.filters.generate_filter
          penalty: mystic.penalty instance of the form y' = k*p(x)
          constraints: mystic.constraints instance of the form x' = c(x)

        NOTE:
          if scipy is not installed, will use np.interp for 1D (non-rbf),
          or mystic's rbf otherwise. default method is 'nearest' for
          1D and 'linear' otherwise. method can be one of ('rbf','linear',
          'nearest','cubic','inverse','gaussian','quintic','thin_plate').
        """
        # sampler configuration
        from mystic.search import Searcher
        self.sampler = Searcher() if sampler is None else sampler
        self.maxpts = kwds.pop('maxpts', None)  # N = 1000
        self.noise = kwds.pop('noise', 1e-8)
        self.filter = kwds.pop('filter', None)
        self.penalty = kwds.pop('penalty', None)
        self.constraints = kwds.pop('constraints', None)
        # monitor, archive, and trajectories
        self._minmon = self._maxmon = None  #XXX: better default?
        self._minarch = self._maxarch = None  #XXX: better default?
        self.x = None  # params (x)
        self.z = None  # cost   (objective(x))
        # point generator(s) and interpolated model(s) #XXX: better names?
        self.dim = kwds.pop('dim', None)  #XXX: should be (or set) len(x)
        self.objective = objective # original F(x)
        self.surrogate = None # interpolated F(*x)
        # interpolator configuration
        self.args = {}#dict(smooth=0, function='thin_plate')
        self.args.update(kwds)
        return

  # XXX: useful?
  # def _invert_model:
  #     takes model and returns inverted_model (maxmodel or invmodel?)
  # def _invert_trajectories:
  #     takes (xyz) trajectories and returns inverted trajectories (xy-z)

    def UseMonitor(self, min=None, max=None):
        """track parameter trajectories with a monitor(s)

        Input:
          min: monitor instance to track minima; if True, use a new Monitor
          max: monitor instance to track maxima; if True, use a new Monitor

        Output:
          None
        """
        from mystic.monitors import Monitor
        if type(min) is bool: self._minmon = Monitor() if min else None
        elif min is not None: self._minmon = min
        if type(max) is bool: self._maxmon = Monitor() if max else None
        elif max is not None: self._maxmon = max
        return


    def UseArchive(self, min=None, max=None):
        """track sampled points in an archive(s)

        Input:
          min: archive instance to store minima; if True, use a new archive
          max: archive instance to store maxima; if True, use a new archive

        Output:
          None
        """
        from klepto.archives import dict_archive as d
        if type(min) is bool: self._minarch = d(cached=False) if min else None
        elif min is not None: self._minarch = min
        if type(max) is bool: self._maxarch = d(cached=False) if max else None
        elif max is not None: self._maxarch = max
        return

    """
    def doit(self, bounds, stop, step=200, scale=False, shift=False,
             density=9, axes=(), vals=(), maxpts=maxpts, **kwds):
        if not self.sampler.traj: self.sampler.UseTrajectories()
        # get trajectories
        self.Sample(bounds, stop)
        # get interpolated function
        self.Interpolate(**kwds)
        # check extrema  #XXX: put _min,_max in Interpolate? (downsampled)
        f = lambda x,z: (z,surface.surrogate(*x))
        print("min: {}; min@f: {}".format(*f(*surface._min())))
        print("max: {}; max@f: {}".format(*f(*urfacef._max())))
        # plot surface
        self.Plot(step=step, scale=scale, shift=shift, density=density, axes=axes, vals=vals, maxpts=maxpts)
        return
    """

    def Sample(self, bounds, stop, clear=False, verbose=False, all=False, **kwds):
        """sample data (x,z) using objective function z=f(x)

        Input:
          bounds: tuple of floats (min,max), bounds on the search region
          stop: termination condition
          clear: if True, clear the archive of stored points
          verbose: if True, print a summary of search/sampling results
          all: if True, use solver EvalMonitor, else use StepMonitor
          filter: a data filter produced with mystic.filters.generate_filter
          penalty: mystic.penalty instance of the form y' = k*p(x)
          constraints: mystic.constraints instance of the form x' = c(x)

        Output:
          x: an array of shape (npts, dim) or (npts,)
          z: an array of shape (npts,)
        """
        #XXX: does the strategy of finding min/max always apply?
        import numpy as np
        penalty = kwds.get('penalty', self.penalty)
        constraints = kwds.get('constraints', self.constraints)
        kwargs = dict(stop=stop, penalty=penalty, constraints=constraints)

        # get model (for minima)
        model = self.objective
        self.dim = len(bounds)

        ### get mins ###
        monitor = self._minmon
        archive = None if clear else self._minarch
        inverse = False

        if all: #FIXME: better define role/use of Reset/Archive/clear...
            self.sampler.Reset(None, inv=inverse) # reset the sampler
            self.sampler.archive = archive
            self.sampler.Search(model, bounds, evalmon=monitor, **kwargs)

        else:
            self.sampler.Reset(archive, inv=inverse) # reset the sampler
            self.sampler.Search(model, bounds, monitor=monitor, **kwargs)
        if verbose: self.sampler._summarize()
        # read trajectories from log (or monitor)
        xyz = self.sampler.Samples(all=all)
        if clear: self.sampler.Reset()  # reset the sampler
        ### end mins ###

        # invert model (for maxima)
        imodel = lambda *args, **kwds: -model(*args, **kwds)
        if penalty is not None: # also invert penalty
            kwargs['penalty'] = lambda *args, **kwds: -penalty(*args, **kwds)

        ### get maxs ###
        monitor = self._maxmon
        archive = None if clear else self._maxarch
        inverse = True

        if all: #FIXME: better define role/use of Reset/Archive/clear...
            self.sampler.Reset(None, inv=inverse) # reset the sampler
            self.sampler.archive = archive
            self.sampler.Search(imodel, bounds, evalmon=monitor, **kwargs)
        else:
            self.sampler.Reset(archive, inv=inverse) # reset the sampler
            self.sampler.Search(imodel, bounds, monitor=monitor, **kwargs)
        if verbose: self.sampler._summarize()
        xyz = np.hstack((xyz, self.sampler.Samples(all=all)))
        if clear: self.sampler.Reset()  # reset the sampler
        ### end maxs ###

        # split into params and cost
        self.x = xyz.T[:,:-1]
        self.z = xyz.T[:,-1]
        # apply any filter, and return
        filter = kwds.pop('filter', self.filter)
        if filter: #XXX: better here, or in Interpolate???
            self.x, self.z = filter(self.x, self.z)
        return self.x, self.z


    def Interpolate(self, **kwds): #XXX: refactor so use self.interpolator ?
        """interpolate data (x,z) to generate response function z=f(*x)

        Input:
          maxpts: int, maximum number of points to use from (x,z)
          noise: float, amplitude of gaussian noise to remove duplicate x
          method: string for kind of interpolator
          extrap: if True, extrapolate a bounding box (can reduce # of nans)
          arrays: if True, return a numpy array; otherwise don't return arrays

        Output:
          interpolated response function, where z=f(*x.T)

        NOTE:
          if scipy is not installed, will use np.interp for 1D (non-rbf),
          or mystic's rbf otherwise. default method is 'nearest' for
          1D and 'linear' otherwise. method can be one of ('rbf','linear',
          'nearest','cubic','inverse','gaussian','quintic','thin_plate').
        """
        from interpolator import Interpolator
        args = self.args.copy()
        args.update(kwds)
        maxpts, noise = self.maxpts, self.noise
        ii = Interpolator(self.x, self.z, maxpts=maxpts, noise=noise, **args)
        self.surrogate = ii.Interpolate(**args)
        # build the surrogate
        self.surrogate.__doc__ = self.objective.__doc__
        return self.surrogate


    def _max(self): #XXX: remove?
        """get the x[i],z[i] corresponding to the max(z)
        """
        import numpy as np
        mz = np.argmax(self.z)
        return self.x[mz], self.z[mz]

    def _min(self): #XXX: remove?
        """get the x[i],z[i] corresponding to the min(z)
        """
        import numpy as np
        mz = np.argmin(self.z)
        return self.x[mz], self.z[mz]


    def Plot(self, **kwds):
        """produce a scatterplot of (x,z) and the surface z = function(*x.T)

        Input:
          step: int, plot every 'step' points on the grid [default: 200]
          scale: float, scaling factor for the z-axis [default: False]
          shift: float, additive shift for the z-axis [default: False]
          density: int, density of wireframe for the plot surface [default: 9]
          axes: tuple, indicies of the axes to plot [default: ()]
          vals: list of values (one per axis) for unplotted axes [default: ()]
          maxpts: int, maximum number of (x,z) points to use [default: None]
          kernel: function transforming x to x', where x' = kernel(x)
          vtol: float, maximum distance outside bounds hypercube to plot data
        """
        # get interpolted function
        fx = self.surrogate
        # plot interpolated surface
        from plotter import Plotter
        p = Plotter(self.x, self.z, fx, **kwds)
        p.Plot()
        # if plotter interpolated the function, get the function
        self.surrogate = fx or p.function


    def __set_function(self, function): #XXX: deal w/ selector (2D)? ExtraArgs?
        # convert to 'model' format (i.e. takes a parameter vector)
        from mystic.math.interpolate import _to_objective
        _objective = _to_objective(function)
        def objective(x, *args, **kwds):
            result = _objective(x, *args, **kwds)
            return result.tolist() if hasattr(result, 'tolist') else result
        self.objective = objective
        self.objective.__doc__ = function.__doc__
        return

    def __function(self): #XXX: deal w/ selector (2D)? ExtraArgs? _to_function
        # convert model to 'args' format (i.e. takes positional args)
        from mystic.math.interpolate import _to_function
        function = _to_function(self.objective, ndim=self.dim)
        function.__doc__ = self.objective.__doc__
        return function

    def __model(self): #XXX: deal w/ selector (2D)? ExtraArgs? _to_objective
        # convert to 'model' format (i.e. takes a parameter vector)
        if self.surrogate is None: return None
        from mystic.math.interpolate import _to_objective
        _objective = _to_objective(self.surrogate)
        def objective(x, *args, **kwds):
            result = _objective(x, *args, **kwds)
            return result.tolist() if hasattr(result, 'tolist') else result
        objective.__doc__ = self.objective.__doc__
        return objective


    # interface
    function = property(__function, __set_function )
    model = property(__model )


# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Example use of Forward Poly Model
in mystic and PARK optimization frameworks.
(built for mystic "trunk" and with park-1.2)

for help, type "python rosetta_parabola_example.py --help"
"""

from math import pi
from numpy import array, real, conjugate
import numpy

try: # check if park is installed
    import park
   #import park.parksnob
    import park.parkde
    Model = park.Model
    __park = True
except ImportError:
    Model = object
    __park = False


def ForwardPolyFactory(params):
    a,b,c = params
    def forward_poly(x):
        """ x should be a 1D (1 by N) numpy array """
        return array((a*x*x + b*x + c))
    return forward_poly

def data(params):
    fwd = ForwardPolyFactory(params)
    x = (array([list(range(101))])-50.)[0]
    return x,fwd(x)


# --- Cost Function stuff ---
# Here is the cost function
def vec_cost_function(params):
    return data(params)[1] - datapts

# Here is the normed version
def cost_function(params):
    x = vec_cost_function(params)
    return numpy.sum(real((conjugate(x)*x)))
# --- Cost Function end ---


# --- Plotting stuff ---
import matplotlib.pyplot as plt
def plot_sol(params,linestyle='b-'):
    d = data(params)
    plt.plot(d[0],d[1],'%s'%linestyle,linewidth=2.0)
    plt.axis(plotview)
    return
# --- Plotting end ---


# --- Call to Mystic ---
def mystic_optimize(point):
    from mystic.monitors import Monitor, VerboseMonitor
    from mystic.solvers import NelderMeadSimplexSolver as fmin
    from mystic.termination import CandidateRelativeTolerance as CRT
    simplex, esow = VerboseMonitor(50), Monitor()
    solver = fmin(len(point))
    solver.SetInitialPoints(point)
    min = [-100,-100,-100]; max = [100,100,100]
    solver.SetStrictRanges(min,max)
    solver.SetEvaluationMonitor(esow)
    solver.SetGenerationMonitor(simplex)
    solver.Solve(cost_function, CRT(1e-7,1e-7))
    solution = solver.Solution()
    return solution
# --- Mystic end ---


# --- Call to Park ---
class PolyModel(Model):
    """a park model:
 - parameters are passed as named strings to set them as class attributes
 - function that does the evaluation must be named "eval"
 - __call__ generated that takes namestring and parameter-named keywords
"""
    parameters = ["a","b","c"]
    def eval(self, x):
        a = self.a
        b = self.b
        c = self.c
        f = ForwardPolyFactory((a,b,c))
        return f(x)
    pass

class Data1D(object):
    """1d model data with the required park functions"""
    def __init__(self,z):
        self.z = z
        return

    def residuals(self,model):
        x = (array([list(range(101))])-50.)[0]
        return (model(x) - self.z).flatten()
    pass


def park_optimize(point):
    # build the data instance
    data1d = Data1D(datapts)

    # build the model instance
    a,b,c = point 
    model = PolyModel("mymodel",a=a,b=b,c=c)
    # required to set bounds on the parameters
    model.a = [-100,100]
    model.b = [-100,100]
    model.c = [-100,100]

    # add a monitor, and set to print results to the console
    handler=park.fitresult.ConsoleUpdate()

    # select the fitter, and do the fit
   #fitter=park.parksnob.Snobfit()
    fitter=park.parkde.DiffEv()
    # 'fit' requires a list of tuples of (model,data)
    result=park.fit.fit([(model,data1d)],fitter=fitter,handler=handler)

    # print results
   #print(result.call)     # print number of function calls
   #result.print_summary() # print solution

    # get the results back into a python object
    solution = {}
    for fitparam in result.parameters:
        solution[fitparam.name] = fitparam.value
    solution = [ solution['mymodel.a'],
                 solution['mymodel.b'],
                 solution['mymodel.c'] ]
    return solution
# --- Park end ---


if __name__ == '__main__':
    # parse user selection to solve with "mystic" [default] or "park"
    from optparse import OptionParser
    parser = OptionParser()
    parser.add_option("-p","--park",action="store_true",dest="park",\
                      default=False,help="solve with park (instead of mystic)")
    parsed_opts, parsed_args = parser.parse_args()

    # set plot window
    from mystic.tools import getch
    plotview = [-10,10, 0,100]

    # Let the "actual parameters" be :
    target = [1., 2., 1.]
    print("Target: %s" % target)

    # Here is the "observed data"
    x,datapts = data(target)
    plt.ion()
    plot_sol(target,'r-')
    plt.draw()
    plt.pause(0.001)

    # initial values
    point = [100,-100,0]

    # DO OPTIMIZATION STUFF HERE TO GET SOLUTION
    if parsed_opts.park:
        if __park:
            print("Solving with park's DE optimizer...")
            solution = park_optimize(point)
        else:
            print('This option requires park to be installed')
            exit()
    else:
        print("Solving with mystic's fmin optimizer...")
        solution = mystic_optimize(point)
    print("Solved: %s" % solution)

    # plot the solution
    plot_sol(solution,'g-')
    plt.draw()
    plt.pause(0.001)

    getch()

# End of file



#!/usr/bin/env python
#
# Problem definition:
# A-R Hedar and M Fukushima, "Derivative-Free Filter Simulated Annealing
# Method for Constrained Continuous Global Optimization", Journal of
# Global Optimization, 35(4), 521-549 (2006).
# 
# Original Matlab code written by A. Hedar (Nov. 23, 2005)
# http://www-optima.amp.i.kyoto-u.ac.jp/member/student/hedar/Hedar_files/go.htm
# and ported to Python by Mike McKerns (December 2014)
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

from g10 import objective, bounds, xs, ys

from mystic.constraints import as_constraint
from mystic.penalty import quadratic_inequality

def penalty1(x): # <= 0.0
    return -1 + 0.0025*(x[3] + x[5])

def penalty2(x): # <= 0.0
    return -1 + 0.0025*(-x[3] + x[4] + x[6])

def penalty3(x): # <= 0.0
    return -1 + 0.01*(-x[4] + x[7])

def penalty4(x): # <= 0.0
    return 100*x[0] - x[0]*x[5] + 833.33252*x[3] - 83333.333

def penalty5(x): # <= 0.0
    return x[1]*x[3] - x[1]*x[6] - 1250*x[3] + 1250*x[4]

def penalty6(x): # <= 0.0
    return x[2]*x[4] - x[2]*x[7] - 2500*x[4] + 1250000

@quadratic_inequality(penalty1, k=1e12)
@quadratic_inequality(penalty2, k=1e12)
@quadratic_inequality(penalty3, k=1e12)
@quadratic_inequality(penalty4, k=1e12)
@quadratic_inequality(penalty5, k=1e12)
@quadratic_inequality(penalty6, k=1e12)
def penalty(x):
    return 0.0

solver = as_constraint(penalty)



if __name__ == '__main__':

    from mystic.solvers import diffev2
    from mystic.math import almostEqual

    result = diffev2(objective, x0=bounds, bounds=bounds, penalty=penalty, npop=80, gtol=500, disp=False, full_output=True)

    assert almostEqual(result[0], xs, rel=1e-2)
    assert almostEqual(result[1], ys, rel=1e-2)



# EOF



#!/usr/bin/env python
#
# Problem definition:
# Example in reference documentation for scipy.optimze
# http://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.curve_fit.html
# 
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
    Fit parameters to noisy data:
               y(x) ~ a * exp(-b * x) + c

    where:
               0 >= x >= 4
               y(x) = y0(x) + yn
               y0(x) = a0 * exp(-b0 * x) + c0
               a0,b0,c0 = 2.5,1.3,0.5
               yn = 0.2 * Normal(0,1)
"""
from numpy import exp, linspace
from numpy.random import normal
from mystic import reduced

def y0(coeffs, x):
    a,b,c = coeffs
    return a * exp(-b * x) + c

coeffs = (2.5, 1.3, 0.5)

# Create noisy data from the 'solution' parameters
x = linspace(0, 4, 50)
y = y0(coeffs, x) + 0.2 * normal(size=len(x))

@reduced(lambda x,y: abs(x)+abs(y))
def objective(coeffs, x, y):
    return y0(coeffs, x) - y

bounds = [(0, 10),(0, 10),(0, 10)]
args = (x,y)

# 'solution' is:
try:
    from scipy.optimize import curve_fit
    xs,pcov = curve_fit(lambda x,*coeffs: y0(coeffs,x), x, y, p0=[1,1,1])
except ImportError:
    xs = x0
ys = objective(xs, x, y)


if __name__ == '__main__':

  from mystic.solvers import diffev2
  from mystic.math import almostEqual
# from mystic.monitors import VerboseMonitor
# mon = VerboseMonitor(10)

  result = diffev2(objective, args=args, x0=bounds, bounds=bounds, npop=40, ftol=1e-8, gtol=100, disp=False, full_output=True)#, itermon=mon)
# print("%s %s" % (result[0], xs))
  assert almostEqual(result[0], xs, rel=2e-1)
  assert almostEqual(result[1], ys, rel=2e-1)

#XXX: how approximate the covariance matrix of estimates (pcov) w/ mystic?
#XXX: mystic should have leastsq


# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @uqfoundation)
# Copyright (c) 2020-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
OUQ classes for calculating bounds on statistical quantities
"""
import mystic.cache
from mystic.cache import cached
from mystic.cache.archive import dict_archive
from mystic.math import almostEqual
from mystic.math.discrete import product_measure
from mystic.math.samples import random_samples
from mystic.solvers import DifferentialEvolutionSolver2
from mystic.monitors import Monitor
from ouq_models import WrapModel


class BaseOUQ(object): #XXX: redo with a "Solver" interface, like ensemble?

    def __init__(self, model, bounds, **kwds):
        """OUQ model for a statistical quantity

    Input:
        model: function of the form y = model(x, axis=None)
        bounds: mystic.bounds.MeasureBounds instance

    Additional Input:
        samples: int, number of samples (used for non-deterministic models)
        constraint: function of the form x' = constraint(x)
        xvalid: function returning True if x == x', given constraint
        cvalid: function similar to xvalid, but with product_measure input

    NOTE: when y is multivalued models must be a UQModel with ny != None
        """
        self.npts = bounds.n  # (2,1,1)
        self.lb = bounds.lower
        self.ub = bounds.upper
        #NOTE: for *_bound(axis=None) to work, requires ny != None
        #if not isinstance(model, WrapModel.mro()[-2]):
        #    model = WrapModel(model=model)
        self.model = model
        self.axes = getattr(model, 'ny', None) #FIXME: in kwds?, ny
        rnd = getattr(model, 'rnd', True) #FIXME: ?, rnd
        self.samples = kwds.get('samples', None) if rnd else None
        self.constraint = kwds.get('constraint', lambda rv:rv)
        self.xvalid = kwds.get('xvalid', lambda rv:True)
        self.cvalid = kwds.get('cvalid', lambda c:True)
        self.kwds = {} #FIXME: good idea???
        self._upper = {} # saved solver instances for upper bounds #XXX: don't?
        self._lower = {} # saved solver instances for lower bounds #XXX: don't?
        self._pts = {} # saved sampled {in:out} objective #XXX: out only? ''?
        return

    # --- extrema ---
    #XXX: expected?

    def upper_bound(self, axis=None, **kwds):
        """find the upper bound on the statistical quantity

    Input:
        axis: int, the index of y on which to find bound (all, by default)
        instance: bool, if True, return the solver instance (False, by default)

    Additional Input:
        kwds: dict, with updates to the instance's stored kwds

    Returns:
        upper bound on the statistical quantity, for the specified axis
        """
        full = kwds.pop('instance', False)
        #self._upper.clear() #XXX: good idea?
        if self.axes is None or axis is not None:
            # solve for upper bound of objective (in measure space)
            solver = self._upper_bound(axis, **kwds)
            self._upper[None if self.axes is None else axis] = solver
            if full: return solver
            return solver.bestEnergy
        # else axis is None
        solvers = self._upper_bound(axis, **kwds)
        for i,solver in enumerate(solvers):
            self._upper[i] = solver
        if full: return solvers
        return tuple(i.bestEnergy for i in solvers)

    def _upper_bound(self, axis=None, **kwds):
        """find the upper bound on the statistical quantity

    Input:
        axis: int, the index of y on which to find bound (all, by default)

    Additional Input:
        kwds: dict, with updates to the instance's stored kwds

    Returns:
        solver instance with solved upper bound on the statistical quantity
        """
        self.kwds.update(**kwds) #FIXME: good idea???
        if self.axes is None or axis is not None:
            # solve for upper bound of objective (in measure space)
            return self.solve(lambda rv: -self.objective(rv, axis))
        # else axis is None
        import multiprocess.dummy as mp #FIXME: process pickle/recursion Error
        pool = mp.Pool(self.axes)
        map = pool.map #TODO: don't hardwire map
        upper = tuple(map(self._upper_bound, range(self.axes)))
        pool.close(); pool.join()
        return upper #FIXME: don't accept "uphill" moves?

    def lower_bound(self, axis=None, **kwds):
        """find the lower bound on the statistical quantity

    Input:
        axis: int, the index of y on which to find bound (all, by default)
        instance: bool, if True, return the solver instance (False, by default)

    Additional Input:
        kwds: dict, with updates to the instance's stored kwds

    Returns:
        lower bound on the statistical quantity, for the specified axis
        """
        full = kwds.pop('instance', False)
        #self._lower.clear() #XXX: good idea?
        if self.axes is None or axis is not None:
            # solve for lower bound of objective (in measure space)
            solver = self._lower_bound(axis, **kwds)
            self._lower[None if self.axes is None else axis] = solver
            if full: return solver
            return solver.bestEnergy
        # else axis is None
        solvers = self._lower_bound(axis, **kwds)
        for i,solver in enumerate(solvers):
            self._lower[i] = solver
        if full: return solvers
        return tuple(i.bestEnergy for i in solvers)

    def _lower_bound(self, axis=None, **kwds):
        """find the lower bound on the statistical quantity

    Input:
        axis: int, the index of y on which to find bound (all, by default)

    Additional Input:
        kwds: dict, with updates to the instance's stored kwds

    Returns:
        solver instance with solved lower bound on the statistical quantity
        """
        self.kwds.update(**kwds) #FIXME: good idea???
        if self.axes is None or axis is not None:
            # solve for lower bound of objective (in measure space)
            return self.solve(lambda rv: self.objective(rv, axis))
        # else axis is None
        import multiprocess.dummy as mp #FIXME: process pickle/recursion Error
        pool = mp.Pool(self.axes)
        map = pool.map #TODO: don't hardwire map
        lower = tuple(map(self._lower_bound, range(self.axes)))
        pool.close(); pool.join()
        return lower #FIXME: don't accept "uphill" moves?

    # --- func ---
    def objective(self, rv, axis=None):
        """calculate the statistical quantity, under uncertainty

    Input:
        rv: list of input parameters
        axis: int, the index of output to calculate (all, by default)

    Returns:
        the statistical quantity for the specified axis

    NOTE:
        respects constraints on input parameters and product measure
        """
        return NotImplemented

    def __call__(self, axis=None, **kwds):
        """apply the reducer to the sampled statistical quantity

    Input:
        axis: int, the index of y on which to find bound (all, by default)

    Additional Input:
        reducer: function to reduce a list to a single value (e.g. mean, max)
        dist: a mystic.tools.Distribution instance (or list of Distributions)
        npts: number of sample points [default = 10000]
        clip: if True, clip at bounds, else resample [default = False]

    Returns:
        sampled statistical quantity, for the specified axis, reduced to a float
        """
        #XXX: return what? "energy and solution?" reduced?
        reducer = kwds.pop('reducer', None)
        if kwds.get('npts', None) is None: kwds.pop('npts', None)
        s = random_samples(self.lb, self.ub, **kwds).T
        fobj = cached(archive=dict_archive())(self.objective) #XXX: bad idea?
        self._pts = fobj.__cache__() #XXX: also bad idea? include from *_bounds?
        objective = lambda rv: fobj(self.constraint(rv), axis=axis)
        import multiprocess.dummy as mp #FIXME: process pickle/recursion Error
        pool = mp.Pool() # len(s)
        map = pool.map #TODO: don't hardwire map
        s = map(objective, s) #NOTE: s = [(...),(...)] or [...]
        pool.close(); pool.join()
        if axis is None and self.axes is not None: # apply per axis
            if reducer is None:
                return tuple(sum(si)/len(si) for si in zip(*s))
            #XXX: better tuple(reducer(s, axis=0).tolist()) if numpy ufunc?
            return tuple(reducer(si) for si in zip(*s))
        s = tuple(s)
        return sum(s)/len(s) if reducer is None else reducer(s)

    # --- solve ---
    def solve(self, objective, **kwds): #NOTE: single axis only
        """solve (in measure space) for bound on given objective

    Input:
        objective: cost function of the form y = objective(x)

    Additional Input:
        solver: mystic.solver instance [default: DifferentialEvolutionSolver2]
        npop: population size [default: None]
        id: a unique identifier for the solver [default: None]
        nested: mystic.solver instance [default: None], for ensemble solvers
        x0: initial parameter guess [default: use RandomInitialPoints]
        pool: pathos.pool instance [default: None]
        maxiter: max number of iterations [default: defined in solver]
        maxfun: max number of objective evaluations [default: defined in solver]
        evalmon: mystic.monitor instance [default: Monitor], for evaluations
        stepmon: mystic.monitor instance [default: Monitor], for iterations
        opts: dict of configuration options for solver.Solve [default: {}]

    Returns:
        solver instance, after Solve has been called
        """
        kwds.update(self.kwds) #FIXME: good idea??? [bad in parallel???]
        lb, ub = self.lb, self.ub
        solver = kwds.get('solver', DifferentialEvolutionSolver2)
        npop = kwds.get('npop', None)
        if npop is not None:
            solver = solver(len(lb),npop)
        else:
            solver = solver(len(lb))
        solver.id = kwds.pop('id', None)
        nested = kwds.get('nested', None)
        x0 = kwds.get('x0', None)
        if nested is not None: # Buckshot/Sparsity
            solver.SetNestedSolver(nested)
        else: # DiffEv/Nelder/Powell
            if x0 is None: solver.SetRandomInitialPoints(min=lb,max=ub)
            else: solver.SetInitialPoints(x0)
        mapper = kwds.get('pool', None)
        if mapper is not None:
            pool = mapper() #XXX: ThreadPool, ProcessPool, etc
            solver.SetMapper(pool.map) #NOTE: not Nelder/Powell
        maxiter = kwds.get('maxiter', None)
        maxfun = kwds.get('maxfun', None)
        solver.SetEvaluationLimits(maxiter,maxfun)
        evalmon = kwds.get('evalmon', None)
        evalmon = Monitor() if evalmon is None else evalmon
        solver.SetEvaluationMonitor(evalmon[:0])
        stepmon = kwds.get('stepmon', None)
        stepmon = Monitor() if stepmon is None else stepmon
        solver.SetGenerationMonitor(stepmon[:0])
        solver.SetStrictRanges(min=lb,max=ub)
        solver.SetConstraints(self.constraint)
        opts = kwds.get('opts', {})
        # solve
        solver.Solve(objective, **opts)
        if mapper is not None:
            pool.close()
            pool.join()
            pool.clear() #NOTE: if used, then shut down pool
        #NOTE: debugging code
        #print("solved: %s" % solver.Solution())
        #func_bound = solver.bestEnergy
        #func_evals = solver.evaluations
        #from mystic.munge import write_support_file
        #write_support_file(solver._stepmon)
        #print("func_bound: %s" % func_bound) #NOTE: may be inverted
        #print("func_evals: %s" % func_evals)
        return solver


class ExpectedValue(BaseOUQ):

    def objective(self, rv, axis=None):
        """calculate expected value of model, under uncertainty

    Input:
        rv: list of input parameters
        axis: int, the index of output to calculate (all, by default)

    Returns:
        the expected value for the specified axis

    NOTE:
        respects constraints on input parameters and product measure

    NOTE:
        for product_measure, use sampled_expect if samples, else expect
        """
        # check constraints
        c = product_measure().load(rv, self.npts)
        if not self.cvalid(c) or not self.xvalid(rv):
            import numpy as np
            return np.inf
        # get expected value
        if axis is None and self.axes is not None:
            model = (lambda x: self.model(x, axis=i) for i in range(self.axes))
            if self.samples is None:
                return tuple(c.expect(m) for m in model)
            # else use sampled support
            return tuple(c.sampled_expect(m, self.samples) for m in model)
        # else, get expected value for the given axis
        if axis is None:
            model = lambda x: self.model(x)
        else:
            model = lambda x: self.model(x, axis=axis)
        if self.samples is None:
            return c.expect(model)
        return c.sampled_expect(model, self.samples)


class MaximumValue(BaseOUQ):

    def objective(self, rv, axis=None):
        """calculate maximum value of model, under uncertainty

    Input:
        rv: list of input parameters
        axis: int, the index of output to calculate (all, by default)

    Returns:
        the maximum value for the specified axis

    NOTE:
        respects constraints on input parameters and product measure

    NOTE:
        for product_measure, use sampled_maximum if samples, else ess_maximum
        """
        # check constraints
        c = product_measure().load(rv, self.npts)
        if not self.cvalid(c) or not self.xvalid(rv):
            import numpy as np
            return np.inf
        # get maximum value
        if axis is None and self.axes is not None:
            model = (lambda x: self.model(x, axis=i) for i in range(self.axes))
            if self.samples is None:
                return tuple(c.ess_maximum(m) for m in model)
            # else use sampled support
            return tuple(c.sampled_maximum(m, self.samples) for m in model)
        # else, get maximum value for the given axis
        if axis is None:
            model = lambda x: self.model(x)
        else:
            model = lambda x: self.model(x, axis=axis)
        if self.samples is None:
            from mystic.math.measures import ess_maximum #TODO: c.ess_maximum
            return ess_maximum(model, c.positions, c.weights)
        return c.sampled_maximum(model, self.samples)


class MinimumValue(BaseOUQ):

    def objective(self, rv, axis=None):
        """calculate minimum value of model, under uncertainty

    Input:
        rv: list of input parameters
        axis: int, the index of output to calculate (all, by default)

    Returns:
        the minimum value for the specified axis

    NOTE:
        respects constraints on input parameters and product measure

    NOTE:
        for product_measure, use sampled_minimum if samples, else ess_minimum
        """
        # check constraints
        c = product_measure().load(rv, self.npts)
        if not self.cvalid(c) or not self.xvalid(rv):
            import numpy as np
            return np.inf
        # get minimum value
        if axis is None and self.axes is not None:
            model = (lambda x: self.model(x, axis=i) for i in range(self.axes))
            if self.samples is None:
                return tuple(c.ess_minimum(m) for m in model)
            # else use sampled support
            return tuple(c.sampled_minimum(m, self.samples) for m in model)
        # else, get minimum value for the given axis
        if axis is None:
            model = lambda x: self.model(x)
        else:
            model = lambda x: self.model(x, axis=axis)
        if self.samples is None:
            from mystic.math.measures import ess_minimum #TODO: c.ess_minimum
            return ess_minimum(model, c.positions, c.weights)
        return c.sampled_minimum(model, self.samples)


class ValueAtRisk(BaseOUQ):

    def objective(self, rv, axis=None):
        """calculate value at risk of model, under uncertainty

    Input:
        rv: list of input parameters
        axis: int, the index of output to calculate (all, by default)

    Returns:
        the value at risk for the specified axis

    NOTE:
        respects constraints on input parameters and product measure

    NOTE:
        for product_measure, use sampled_ptp if samples, else ess_ptp
        """
        # check constraints
        c = product_measure().load(rv, self.npts)
        if not self.cvalid(c) or not self.xvalid(rv):
            import numpy as np
            return np.inf
        # get value at risk
        if axis is None and self.axes is not None:
            model = (lambda x: self.model(x, axis=i) for i in range(self.axes))
            if self.samples is None:
                return tuple(c.ess_ptp(m) for m in model)
            # else use sampled support
            return tuple(c.sampled_ptp(m, self.samples) for m in model)
        # else, get value at risk for the given axis
        if axis is None:
            model = lambda x: self.model(x)
        else:
            model = lambda x: self.model(x, axis=axis)
        if self.samples is None:
            from mystic.math.measures import ess_ptp #TODO: c.ess_ptp
            return ess_ptp(model, c.positions, c.weights)
        return c.sampled_ptp(model, self.samples)


class ProbOfFailure(BaseOUQ):

    def objective(self, rv, axis=None, iter=True):
        """calculate probability of failure for model, under uncertainty

    Input:
        rv: list of input parameters
        axis: int, the index of output to calculate (all, by default)
        iter: bool, if True, calculate per axis, else calculate for all axes

    Returns:
        the probability of failure for the specified axis (or all axes)

    NOTE:
        respects constraints on input parameters and product measure
        model is a function returning a boolean (True for success)

    NOTE:
        for product_measure, use sampled_pof(model) if samples, else pof(model)
        """
        # check constraints
        c = product_measure().load(rv, self.npts)
        if not self.cvalid(c) or not self.xvalid(rv):
            import numpy as np
            return np.inf
        # get probability of failure
        if iter and axis is None and self.axes is not None:
            model = (lambda x: self.model(x,axis=i) for i in range(self.axes))
            if self.samples is None:
                return tuple(c.pof(m) for m in model)
            # else use sampled support
            return tuple(c.sampled_pof(m, self.samples) for m in model)
        # else, get probability of failure for the given axis
        if axis is None:
            model = lambda x: self.model(x)
        else:
            model = lambda x: self.model(x, axis=axis)
        if self.samples is None:
            return c.pof(model)
        return c.sampled_pof(model, self.samples)




#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2020-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
  Example applying mystic to sklearn

  Use a linear regression to fit sparse data generated from:
            f(x) = a*x3**3 + b*x2**2 + c*x1 + d*x0
            a,b,c,d = 0.661, -1.234, 2.983, -16.5571

  Where the following information is utilized: 
            f(x) is a polynomial of order=3
            3*b + c > -0.75
            4.5*b - d > 11.0
"""
import numpy as np
from sklearn import preprocessing as pre
from sklearn import linear_model as lin
from mystic.symbolic import generate_constraint, generate_solvers, simplify
from mystic.constraints import vectorize
from mystic import random_seed
random_seed(123)

# define a model
a,b,c,d = 0.661, -1.234, 2.983, -16.5571
def model(x):
  x0,x1,x2,x3 = x
  return a*x3**3 + b*x2**2 + c*x1 + d*x0

# generate some sparse data
xtrain = np.random.uniform(0,100, size=(10,4))
target = model(xtrain.T).T
xtest = np.random.uniform(0,100, size=(10,4))
test = model(xtest.T).T

# define some model constraints
equations = """
3*b + c > -0.75
4.5*b - d > 11.0
"""
var = list('abcd')
equations = simplify(equations, variables=var)
cf = generate_constraint(generate_solvers(equations, variables=var))


if __name__ == '__main__':
    # build a kernel-transformed regressor
    ta = pre.FunctionTransformer(func=vectorize(cf, axis=1))
    tp = pre.PolynomialFeatures(degree=3)
    e = lin.LinearRegression()

    # train and score, then test and score
    xtrain_ = tp.fit_transform(ta.fit_transform(xtrain))
    assert 1.0 == e.fit(xtrain_, target).score(xtrain_, target)
    xtest_ = tp.fit_transform(ta.fit_transform(xtest))
    assert 1 - e.score(xtest_, test) <= 1e-2

# EOF



#!/usr/bin/env python
#
# Problem definition:
# Example in reference documentation for cvxopt
# http://cvxopt.org/examples/tutorial/qp.html
# 
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
    Minimize: f = 2*x[0]**2 + x[1]**2 + x[0]*x[1] + x[0] + x[1]

    Subject to:     x[0] >= 0
                           x[1] >= 0
                    x[0] + x[1] == 1
"""

def objective(x):
  x0,x1 = x
  return 2*x0**2 + x1**2 + x0*x1 + x0 + x1

equations = """
x0 + x1 - 1.0 == 0.0
"""
bounds = [(0.0, None),(0.0, None)]

# with penalty='penalty' applied, solution is:
xs = [0.25, 0.75]
ys = 1.875

from mystic.symbolic import generate_conditions, generate_penalty
pf = generate_penalty(generate_conditions(equations), k=1e4)
from mystic.symbolic import generate_constraint, generate_solvers, solve
cf = generate_constraint(generate_solvers(solve(equations)))


if __name__ == '__main__':

  from mystic.solvers import diffev2, fmin_powell
  from mystic.math import almostEqual

  result = diffev2(objective, x0=bounds, bounds=bounds, constraint=cf, penalty=pf, npop=40, disp=False, full_output=True)
  assert almostEqual(result[0], xs, rel=2e-2)
  assert almostEqual(result[1], ys, rel=2e-2)

  result = fmin_powell(objective, x0=[0.0,0.0], bounds=bounds, constraint=cf, penalty=pf, disp=False, full_output=True)
  assert almostEqual(result[0], xs, rel=2e-2)
  assert almostEqual(result[1], ys, rel=2e-2)


# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Example:
    - Minimize Rosenbrock's Function with Nelder-Mead.
    - Dynamic plot of parameter convergence to function minimum.

Demonstrates:
    - standard models
    - minimal solver interface
    - parameter trajectories using callback
    - solver interactivity
"""

# Nelder-Mead solver
from mystic.solvers import fmin

# Rosenbrock function
from mystic.models import rosen

# tools
from mystic.tools import getch
import matplotlib.pyplot as plt
plt.ion()

# draw the plot
def plot_frame():
    plt.title("Rosenbrock parameter convergence")
    plt.xlabel("Nelder-Mead solver iterations")
    plt.ylabel("parameter value")
    plt.draw()
    plt.pause(0.001)
    return
 
iter = 0
step, xval, yval, zval = [], [], [], []
# plot the parameter trajectories
def plot_params(params):
    global iter, step, xval, yval, zval
    step.append(iter)
    xval.append(params[0])
    yval.append(params[1])
    zval.append(params[2])
    plt.plot(step,xval,'b-')
    plt.plot(step,yval,'g-')
    plt.plot(step,zval,'r-')
    plt.legend(["x", "y", "z"])
    plt.draw()
    plt.pause(0.001)
    iter += 1
    return


if __name__ == '__main__':

    # initial guess
    x0 = [0.8,1.2,0.7]

    # suggest that the user interacts with the solver
    print("NOTE: while solver is running, press 'Ctrl-C' in console window")
    getch()
    plot_frame()

    # use Nelder-Mead to minimize the Rosenbrock function
    solution = fmin(rosen,x0,disp=1,callback=plot_params,handler=True)
    print(solution)

    # don't exit until user is ready
    getch()

# end of file



#!/usr/bin/env python
#
# Problem definition:
# A-R Hedar and M Fukushima, "Derivative-Free Filter Simulated Annealing
# Method for Constrained Continuous Global Optimization", Journal of
# Global Optimization, 35(4), 521-549 (2006).
# 
# Original Matlab code written by A. Hedar (Nov. 23, 2005)
# http://www-optima.amp.i.kyoto-u.ac.jp/member/student/hedar/Hedar_files/go.htm
# and ported to Python by Mike McKerns (December 2014)
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

def objective(x):
    from numpy import exp, product
    return exp(product(x))

bounds = [(-2.3,2.3)]*2 + [(-3.2,3.2)]*3
# with penalty='penalty' applied, solution is:
xs = [-1.717143, 1.595709, 1.827247, -0.7636413, -0.763645]
_xs = [-1.717143, 1.595709, 1.827247, 0.7636413, 0.763645]
x_s = [-1.717143, 1.595709, -1.827247, 0.7636413, -0.763645]
xs_ = [-1.717143, 1.595709, -1.827247, -0.7636413, 0.763645]
ys = 0.05394983

from mystic.symbolic import generate_constraint, generate_solvers, simplify
from mystic.symbolic import generate_penalty, generate_conditions

equations = """
x0**2 + x1**2 + x2**2 + x3**2 + x4**2 - 10.0 = 0.0
x1*x2 - 5.0*x3*x4 = 0.0
x0**3 + x1**3 + 1.0 = 0.0
"""
cf = generate_constraint(generate_solvers(simplify(equations))) # slow solve
pf = generate_penalty(generate_conditions(equations))



if __name__ == '__main__':

    from mystic.solvers import lattice
    from mystic.math import almostEqual

    result = lattice(objective, 5, [2]*5, bounds=bounds, penalty=pf, ftol=1e-8, xtol=1e-8, disp=False, full_output=True)

    assert almostEqual(result[0], xs, tol=5e-2) \
        or almostEqual(result[0], _xs, tol=5e-2) \
        or almostEqual(result[0], x_s, tol=5e-2) \
        or almostEqual(result[0], xs_, tol=5e-2)
    assert almostEqual(result[1], ys, rel=5e-2)

# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

from mystic.tools import list_or_tuple_or_ndarray as sequence
from mystic.tools import isNull

import sys
if (sys.hexversion >= 0x30000f0):
    exec_string = 'exec(code, globals)'
else:
    exec_string = 'exec code in globals'

# generalized history reader

def read_history(source):
    """read parameter history and cost history from the given source

source is a monitor, logfile, support file, solver restart file, dataset, etc
    """
    monitor = solver = data = arxiv = False
    from mystic.monitors import Monitor, Null
    from mystic.abstract_solver import AbstractSolver
    from mystic.math.legacydata import dataset
    from klepto.archives import archive, cache
    try:
        basestring
    except NameError:
        basestring = str
        import io
        file = io.IOBase
    if isinstance(source, file):
        return read_history(source.name)
    if isinstance(source, basestring):
        import re
        source = re.sub(r'\.py*.$', '', source)  # strip off .py* extension
    elif isinstance(source, Monitor):
        monitor = True
    elif isinstance(source, dataset):
        data = True
    elif isinstance(source, cache):
        source = source.__archive__
        arxiv = True
    elif isinstance(source, archive):
        arxiv = True
    elif isinstance(source, AbstractSolver):
        solver = True
    elif isinstance(source, Null):
        return [],[] #XXX: or source.x, source.y (e.g. Null(),Null())? or Error?
    else:
        raise IOError("a history filename or instance is required")
    try:  # read standard logfile (or monitor)
        if monitor:
            params, cost = read_monitor(source)
        elif data:
            params, cost = source.coords, source.values
        elif arxiv:
            params = list(list(k) for k in source.keys())
            cost = source.values()
            cost = cost if type(cost) is list else list(cost)
        elif solver:
            params, cost = source.solution_history, source.energy_history
        else:
            try:
                from mystic.solvers import LoadSolver
                return read_history(LoadSolver(source))
            except: #KeyError
                _step, params, cost = logfile_reader(source)
                #FIXME: doesn't work for multi-id logfile; select id?
        params, cost = raw_to_support(params, cost)
    except:
        params, cost = read_import(source+'.py', 'params', 'cost')
    return params, cost


# logfile reader
#XXX: if have grad ([i,j,k]) not cost, 'cost' is list of lists [[i,j,k],...]
#XXX: if have grad and cost, 'cost' is mixed list [n,[i,j,k],m,[a,b,c],...]
#TODO: (mixed) create a function that splits/filters list/float in 'cost'
#TODO: (mixed) split/filter values in 'step','param' using indices from 'cost'
#TODO: behavior is identical for 'raw_to_support'... apply changes there also

def logfile_reader(filename):
  from numpy import inf, nan
  f = open(filename,"r")
  file = f.read()
  f.close()
  contents = file.split("\n")
  # parse file contents to get (i,id), cost, and parameters
  step = []; cost = []; param = [];
  for line in contents[:-1]:
    if line.startswith("#"): pass
    else:
      values = line.split("   ")
      step.append(eval(values[0]))  #XXX: yields (i,id)
      cost.append(eval(values[1]))
      param.append(eval(values[2]))
  return step, param, cost

def read_trajectories(source):
  """read trajectories from a convergence logfile or a monitor

source can either be a monitor instance or a logfile path
  """
  if isinstance(source, str):
    step, param, cost = logfile_reader(source)
  else:
    if source.id and isinstance(source.id[0], tuple):
      step = source.id
    else:
      step = enumerate(source.id)
    if len(source) == source.id.count(None):
      step = [(i,) for (i,j) in step]
    else:
      step = list(step)
    param, cost = source.x, source.y
  return step, param, cost


# read and write monitor (to and from raw data)

def read_monitor(mon, id=False):
  steps = mon.x[:]
  energy = mon.y[:]
  if not id:
    return steps, energy
  id = mon.id[:]
  return steps, energy, id 

def write_monitor(steps, energy, id=[], k=None):
  from mystic.monitors import Monitor
  mon = Monitor()
  mon.k = k
  mon._x.extend(steps)
  mon._y.extend(mon._k(energy, iter))
  mon._id.extend(id)
  return mon

# converters 

def converge_to_support(steps, energy):
  steps = zip(*steps)
  steps = [list(i) for i in steps]
  return steps, energy

def raw_to_converge(steps, energy):
  if len(steps) > 0:
    if not sequence(steps[0][0]):
      steps = [[step] for step in steps]  # needed when steps = [1,2,3,...]
    steps = [list(zip(*step)) for step in steps] # also can be used to revert 'steps'
  return steps, energy

def raw_to_support(steps, energy):
  return converge_to_support( *raw_to_converge(steps, energy) )

# monitor to file (support file, converge file, raw file)
## FIXME: 'converge' and 'raw' files are virtually unused and unsupported

def write_raw_file(mon,log_file='paramlog.py',**kwds):
  if isNull(mon): return  #XXX: throw error? warning? ???
  steps, energy = read_monitor(mon)
  f = open(log_file,'w')
  if 'header' in kwds:
    f.write('# %s\n' % kwds['header'])
    del kwds['header']
  f.write("inf = float('inf')\n") # define special values
  f.write("nan = float('nan')\n") # define special values
  for variable,value in getattr(kwds, 'iteritems', kwds.items)():
    f.write('%s = %s\n' % (variable,value))# write remaining kwds as variables
 #f.write('# %s\n' % energy[-1])
  f.write('params = %s\n' % steps)
  f.write('cost = %s\n' % energy)
  f.close()
  return

def write_support_file(mon,log_file='paramlog.py',**kwds):
  if isNull(mon): return  #XXX: throw error? warning? ???
  monitor = write_monitor( *raw_to_support( *read_monitor(mon) ) )
  header = "written in 'support' format"
  if 'header' in kwds:
    header += "\n# " + str(kwds['header'])
    del kwds['header']
  write_raw_file(monitor,log_file,header=header,**kwds)
  return

def write_converge_file(mon,log_file='paramlog.py',**kwds):
  if isNull(mon): return  #XXX: throw error? warning? ???
  monitor = write_monitor( *raw_to_converge( *read_monitor(mon) ) )
  header = "written in 'converge' format"
  if 'header' in kwds:
    header += "\n# " + str(kwds['header'])
    del kwds['header']
  write_raw_file(monitor,log_file,header=header,**kwds)
  return

# file to data (support file, converge file, raw file)

def read_raw_file(file_in):
  steps, energy = read_import(file_in, "params", "cost")
  return steps, energy  # was 'from file_in import params as steps', etc

#TODO: check impact of having gradient ([i,j,k]) and/not cost
def_read_import = r'''
def read_import(file, *targets):
  "import the targets; targets are name strings"
  import re, os, sys
  _dir, file = os.path.split(file)
  file = re.sub(r'\.py*.$', '', file) #XXX: strip .py* extension
  curdir = os.path.abspath(os.curdir)
  sys.path.append('.')
  results = []
  globals = {}
  try:
    if _dir: os.chdir(_dir)
    if len(targets):
      for target in targets:
        code = "from {0} import {1} as result".format(file, target)
        code = compile(code, '<string>', 'exec')
        %(exec_string)s
        results.append(globals['result'])
    else:
        code = "import {0} as result".format(file)
        code = compile(code, '<string>', 'exec')
        %(exec_string)s
        results.append(globals['result'])
  except ImportError:
    raise RuntimeError('File: {0} not found'.format(file))
  finally:
    if _dir: os.chdir(curdir)
    sys.path.pop()
  if not len(results): return None
  return results[-1] if (len(results) == 1) else results
''' % dict(exec_string=exec_string)

exec(def_read_import)
del def_read_import

def read_converge_file(file_in):
  steps, energy = read_raw_file(file_in)
 #steps = [zip(*step) for step in steps] # also can be used to revert 'steps'
  return raw_to_converge(steps,energy)

def read_support_file(file_in):
  steps, energy = read_raw_file(file_in)
  return raw_to_support(steps,energy)
 #return converge_to_support(steps,energy)

# file converters

def raw_to_converge_converter(file_in,file_out):
  steps, energy = read_raw_file(file_in)
  write_raw_file(write_monitor( *raw_to_converge(steps,energy) ),
                 file_out,header="written in 'converge' format")
  return

def raw_to_support_converter(file_in,file_out):
  steps, energy = read_raw_file(file_in)
  write_raw_file(write_monitor( *raw_to_support(steps,energy) ),
                 file_out,header="written in 'support' format")
  return

def converge_to_support_converter(file_in,file_out):
  steps, energy = read_converge_file(file_in)
  write_raw_file(write_monitor( *converge_to_support(steps,energy) ),
                 file_out,header="written in 'support' format")
  return

# old #

def read_old_support_file(file_in):
  steps, energy = read_raw_file(file_in)
  steps = [[(i,) for i in steps[j]] for j in range(len(steps))]
  energy = [(i,) for i in energy]
  return write_monitor(steps, energy)

def old_to_new_support_converter(file_in,file_out):
  mon = read_old_support_file(file_in)
  write_raw_file(mon,file_out)
  return

def __orig_write_support_file(mon,log_file='paramlog.py'):
  if isNull(mon): return  #XXX: throw error? warning? ???
  steps, energy = read_monitor(mon)
  log = []
  if len(steps) > 0:
    for p in range(len(steps[0])):
      q = []
      for s in range(len(steps)):
        q.append(steps[s][p])
      log.append(q)  
  monitor = write_monitor(log, energy)
  write_raw_file(monitor,log_file)
  return

def __orig_write_converge_file(mon,log_file='paramlog.py'):
  write_raw_file(mon,log_file)
  return


if __name__ == '__main__':
  pass

# EOF

